# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2011-2012 Lambda Foundry, Inc. and PyData Development Team
# Copyright (c) 2013 Jev Kuznetsov and contributors
# Copyright (c) 2014-2015 Scott Hansen <firecat4153@gmail.com>
# Copyright (c) 2014-2016 Yuri D'Elia "wave++" <wavexx@thregr.org>
# Copyright (c) 2014- Spyder Project Contributors
#
# Components of gtabview originally distributed under the MIT (Expat) license.
# This file as a whole distributed under the terms of the New BSD License
# (BSD 3-clause; see NOTICE.txt in the Spyder root directory for details).
# -----------------------------------------------------------------------------

"""
Pandas DataFrame Editor Dialog.

DataFrameModel is based on the class ArrayModel from array editor
and the class DataFrameModel from the pandas project.
Present in pandas.sandbox.qtpandas in v0.13.1.

DataFrameHeaderModel and DataFrameLevelModel are based on the classes
Header4ExtModel and Level4ExtModel from the gtabview project.
DataFrameModel is based on the classes ExtDataModel and ExtFrameModel, and
DataFrameEditor is based on gtExtTableView from the same project.

DataFrameModel originally based on pandas/sandbox/qtpandas.py of the
`pandas project <https://github.com/pandas-dev/pandas>`_.
The current version is qtpandas/models/DataFrameModel.py of the
`QtPandas project <https://github.com/draperjames/qtpandas>`_.

Components of gtabview from gtabview/viewer.py and gtabview/models.py of the
`gtabview project <https://github.com/TabViewer/gtabview>`_.
"""

# Standard library imports
from __future__ import annotations
import io
from time import perf_counter
from typing import Any, Callable, Optional, TYPE_CHECKING

# Third party imports
from packaging.version import parse
from qtpy.compat import from_qvariant, to_qvariant
from qtpy.QtCore import (
    QAbstractTableModel, QEvent, QItemSelectionModel, QModelIndex, QPoint, Qt,
    Signal, Slot)
from qtpy.QtGui import QColor, QCursor
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QHeaderView,
    QInputDialog,
    QItemDelegate,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QScrollBar,
    QStyle,
    QTableView,
    QTableWidget,
    QToolButton,
    QVBoxLayout,
    QWidget,
)
from spyder_kernels.utils.lazymodules import numpy as np, pandas as pd

# Local imports
from spyder.api.fonts import SpyderFontsMixin, SpyderFontType
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import running_under_pytest
from spyder.plugins.variableexplorer.widgets.arrayeditor import get_idx_rect
from spyder.plugins.variableexplorer.widgets.basedialog import BaseDialog
from spyder.plugins.variableexplorer.widgets.preferences import (
    PreferencesDialog
)
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.qthelpers import keybinding, qapplication
from spyder.utils.stylesheet import AppStyle, MAC
from spyder.widgets.helperwidgets import MessageCheckBox


if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from pandas import DataFrame
    from spyder.plugins.variableexplorer.widgets.namespacebrowser import (
        NamespaceBrowser
    )


# =============================================================================
# ---- Constants
# =============================================================================

class DataframeEditorActions:
    Close = 'close'
    ConvertToBool = 'convert_to_bool_action'
    ConvertToComplex = 'convert_to_complex_action'
    ConvertToFloat = 'convert_to_float_action'
    ConvertToInt = 'convert_to_int_action'
    ConvertToStr = 'convert_to_str_action'
    Copy = 'copy_action'
    DuplicateColumn = 'duplicate_column_action'
    DuplicateRow = 'duplicate_row_action'
    Edit = 'edit_action'
    EditHeader = 'edit_header_action'
    EditIndex = 'edit_index_action'
    Histogram = 'histogram'
    InsertAbove = 'insert_above_action'
    InsertAfter = 'insert_after_action'
    InsertBefore = 'insert_before_action'
    InsertBelow = 'insert_below_action'
    Preferences = 'preferences_action'
    Refresh = 'refresh_action'
    RemoveColumn = 'remove_column_action'
    RemoveRow = 'remove_row_action'
    ResizeColumns = 'resize_columns_action'
    ResizeRows = 'resize_rows_action'


class DataframeEditorMenus:
    Context = 'context_menu'
    ConvertTo = 'convert_to_submenu'
    Header = 'header_context_menu'
    Index = 'index_context_menu'
    Options = 'options_menu'


class DataframeEditorWidgets:
    OptionsToolButton = 'options_button_widget'
    Toolbar = 'toolbar'
    ToolbarStretcher = 'toolbar_stretcher'


class DataframeEditorContextMenuSections:
    Edit = 'edit_section'
    Row = 'row_section'
    Column = 'column_section'
    Convert = 'convert_section'


class DataframeEditorToolbarSections:
    Row = 'row_section'
    ColumnAndRest = 'column_section'


# Supported real and complex number types
REAL_NUMBER_TYPES = (float, int, np.int64, np.int32)
COMPLEX_NUMBER_TYPES = (complex, np.complex64, np.complex128)

# Used to convert bool intrance to false since bool('False') will return True
_bool_false = ['false', 'f', '0', '0.', '0.0', ' ']

# Default format for data frames with floats
DEFAULT_FORMAT = '.6g'

# Limit at which dataframe is considered so large that it is loaded on demand
LARGE_SIZE = 5e5
LARGE_NROWS = 1e5
LARGE_COLS = 60
ROWS_TO_LOAD = 500
COLS_TO_LOAD = 40

# Background colours
BACKGROUND_NUMBER_MINHUE = 0.66  # hue for largest number
BACKGROUND_NUMBER_HUERANGE = 0.33  # (hue for smallest) minus (hue for largest)
BACKGROUND_NUMBER_SATURATION = 0.7
BACKGROUND_NUMBER_VALUE = 1.0
BACKGROUND_NUMBER_ALPHA = 0.6
BACKGROUND_NONNUMBER_COLOR = SpyderPalette.COLOR_BACKGROUND_2
BACKGROUND_STRING_ALPHA = 0.05
BACKGROUND_MISC_ALPHA = 0.3

# =============================================================================
# ---- Utility functions
# =============================================================================

def is_any_real_numeric_dtype(dtype) -> bool:
    """
    Test whether a Pandas dtype is a real numeric type.
    """
    try:
        import pandas.api.types
        return pandas.api.types.is_any_real_numeric_dtype(dtype)
    except Exception:
        # Pandas version 1
        return dtype in REAL_NUMBER_TYPES


def bool_false_check(value):
    """
    Used to convert bool entrance to false.

    Needed since any string in bool('') will return True.
    """
    if value.lower() in _bool_false:
        value = ''
    return value


def global_max(col_vals, index):
    """Returns the global maximum and minimum."""
    col_vals_without_None = [x for x in col_vals if x is not None]
    max_col, min_col = zip(*col_vals_without_None)
    return max(max_col), min(min_col)


# =============================================================================
# ---- Main classes
# =============================================================================

class DataFrameModel(QAbstractTableModel, SpyderFontsMixin):
    """
    Model encapsulating a dataframe for the datafgrame editor

    This is a model (in the sense of the Qt model/view architecture).

    Partly based in ExtDataModel and ExtFrameModel classes
    of the gtabview project.

    For more information please see:
    https://github.com/wavexx/gtabview/blob/master/gtabview/models.py

    Parameters
    ----------
    dataFrame : DataFrame
        The dataframe encapsulated by the model.
    format_spec : str, optional
        Format specification for floats. The default is DEFAULT_FORMAT.
    parent : Optional[QWidget], optional
        The parent widget for the model. The default is None.
    readonly : bool, optional
        If True, the underlying dataframe can not be edited by the user.
        The default is False.

    Attributes
    ----------
    bgcolor_enabled : bool
        If True, vary backgrond color depending on cell value
    colum_avg_enabled : bool
        If True, select background color by comparing cell value against
        column maximum and minimum. Otherwise, use maximum and minimum of
        the entire dataframe.
    _format_spec : str
        Format specification for floats
    """

    def __init__(
        self,
        dataFrame: DataFrame,
        format_spec: str = DEFAULT_FORMAT,
        parent: Optional[QWidget] = None,
        readonly: bool = False,
    ):
        QAbstractTableModel.__init__(self)
        self.dialog = parent
        self.df = dataFrame
        self.df_columns_list = None
        self.df_index_list = None
        self._format_spec = format_spec
        self.readonly = readonly
        self.complex_intran = None
        self.display_error_idxs = []

        self.total_rows = self.df.shape[0]
        self.total_cols = self.df.shape[1]
        size = self.total_rows * self.total_cols

        self.max_min_col = None
        if size < LARGE_SIZE:
            self.max_min_col_update()
            self.colum_avg_enabled = True
            self.bgcolor_enabled = True
            self.colum_avg(True)
        else:
            self.colum_avg_enabled = False
            self.bgcolor_enabled = False
            self.colum_avg(False)

        # Use paging when the total size, number of rows or number of
        # columns is too large
        if size > LARGE_SIZE:
            self.rows_loaded = ROWS_TO_LOAD
            self.cols_loaded = COLS_TO_LOAD
        else:
            if self.total_rows > LARGE_NROWS:
                self.rows_loaded = ROWS_TO_LOAD
            else:
                self.rows_loaded = self.total_rows
            if self.total_cols > LARGE_COLS:
                self.cols_loaded = COLS_TO_LOAD
            else:
                self.cols_loaded = self.total_cols

    def _axis(self, axis):
        """
        Return the corresponding labels taking into account the axis.

        The axis could be horizontal (0) or vertical (1).
        """
        return self.df.columns if axis == 0 else self.df.index

    def _axis_list(self, axis):
        """
        Return the corresponding labels as a list taking into account the axis.

        The axis could be horizontal (0) or vertical (1).
        """
        if axis == 0:
            if self.df_columns_list is None:
                self.df_columns_list = self.df.columns.tolist()
            return self.df_columns_list
        else:
            if self.df_index_list is None:
                self.df_index_list = self.df.index.tolist()
            return self.df_index_list

    def _axis_levels(self, axis):
        """
        Return the number of levels in the labels taking into account the axis.

        Get the number of levels for the columns (0) or rows (1).
        """
        ax = self._axis(axis)
        return 1 if not hasattr(ax, 'levels') else len(ax.levels)

    @property
    def shape(self):
        """Return the shape of the dataframe."""
        return self.df.shape

    @property
    def header_shape(self):
        """Return the levels for the columns and rows of the dataframe."""
        return (self._axis_levels(0), self._axis_levels(1))

    @property
    def chunk_size(self):
        """Return the max value of the dimensions of the dataframe."""
        return max(*self.shape())

    def header(self, axis, x, level=0):
        """
        Return the values of the labels for the header of columns or rows.

        The value corresponds to the header of column or row x in the
        given level.
        """
        ax = self._axis(axis)
        if not hasattr(ax, 'levels'):
            ax = self._axis_list(axis)
            if len(ax) > 0:
                return ax[x]
            else:
                return None
        else:
            return ax.values[x][level]

    def name(self, axis, level):
        """Return the labels of the levels if any."""
        ax = self._axis(axis)
        if hasattr(ax, 'levels'):
            return ax.names[level]
        if ax.name:
            return ax.name

    def max_min_col_update(self):
        """
        Determines the maximum and minimum number in each column.

        The result is a list whose k-th entry is [vmax, vmin], where vmax and
        vmin denote the maximum and minimum of the k-th column (ignoring NaN).
        This list is stored in self.max_min_col.

        If the k-th column has a non-numerical dtype, then the k-th entry
        is set to None. If the dtype is complex, then compute the maximum and
        minimum of the absolute values. If vmax equals vmin, then vmin is
        decreased by one.
        """
        if self.df.shape[0] == 0:  # If no rows to compute max/min then return
            return
        self.max_min_col = []
        for __, col in self.df.items():
            # This is necessary to catch some errors in Pandas when computing
            # the maximum of a column.
            # Fixes spyder-ide/spyder#17145 and spyder-ide/spyder#24094
            try:
                if (
                    is_any_real_numeric_dtype(col.dtype)
                    or col.dtype in COMPLEX_NUMBER_TYPES
                ):
                    if is_any_real_numeric_dtype(col.dtype):
                        vmax = col.max(skipna=True)
                        vmin = col.min(skipna=True)
                    else:
                        vmax = col.abs().max(skipna=True)
                        vmin = col.abs().min(skipna=True)
                    if vmax != vmin:
                        max_min = [vmax, vmin]
                    else:
                        max_min = [vmax, vmin - 1]
                else:
                    max_min = None
            except (TypeError, ValueError):
                max_min = None

            self.max_min_col.append(max_min)

    def get_format_spec(self) -> str:
        """
        Return current format specification for floats.
        """
        # Avoid accessing the private attribute _format_spec from outside
        return self._format_spec

    def set_format_spec(self, format_spec: str) -> None:
        """
        Set format specification for floats.
        """
        self._format_spec = format_spec
        self.reset()

    def bgcolor(self, value: bool):
        """
        Set whether background color varies depending on cell value.
        """
        self.bgcolor_enabled = value
        self.reset()

    def colum_avg(self, value: bool):
        """
        Set what to compute cell value with to choose background color.

        If `value` is True, then compare against column maximum and minimum,
        otherwise compare against maximum and minimum over all columns.
        """
        self.colum_avg_enabled = value
        if self.colum_avg_enabled:
            self.return_max = lambda col_vals, index: col_vals[index]
        else:
            self.return_max = global_max
        self.reset()

    def get_bgcolor(self, index):
        """Background color depending on value."""
        column = index.column()

        if not self.bgcolor_enabled:
            return

        value = self.get_value(index.row(), column)
        if self.max_min_col[column] is None or pd.isna(value):
            color = QColor(BACKGROUND_NONNUMBER_COLOR)
            if isinstance(value, str):
                color.setAlphaF(BACKGROUND_STRING_ALPHA)
            else:
                color.setAlphaF(BACKGROUND_MISC_ALPHA)
        else:
            if isinstance(value, COMPLEX_NUMBER_TYPES):
                color_func = abs
            else:
                color_func = float
            vmax, vmin = self.return_max(self.max_min_col, column)

            # This is necessary to catch an error in Pandas when computing
            # the difference between the max and min of a column.
            # Fixes spyder-ide/spyder#18005
            try:
                if vmax - vmin == 0:
                    vmax_vmin_diff = 1.0
                else:
                    vmax_vmin_diff = vmax - vmin
            except TypeError:
                return

            hue = (BACKGROUND_NUMBER_MINHUE + BACKGROUND_NUMBER_HUERANGE *
                   (vmax - color_func(value)) / (vmax_vmin_diff))
            hue = float(abs(hue))
            if hue > 1:
                hue = 1
            color = QColor.fromHsvF(hue, BACKGROUND_NUMBER_SATURATION,
                                    BACKGROUND_NUMBER_VALUE,
                                    BACKGROUND_NUMBER_ALPHA)

        return color

    def get_value(self, row, column):
        """Return the value of the DataFrame."""
        # To increase the performance iat is used but that requires error
        # handling, so fallback uses iloc
        try:
            value = self.df.iat[row, column]
        except pd._libs.tslib.OutOfBoundsDatetime:
            value = self.df.iloc[:, column].astype(str).iat[row]
        except:
            value = self.df.iloc[row, column]
        return value

    def data(self, index, role=Qt.DisplayRole):
        """Cell content"""
        if not index.isValid():
            return to_qvariant()
        if role == Qt.DisplayRole or role == Qt.EditRole:
            column = index.column()
            row = index.row()
            value = self.get_value(row, column)
            if isinstance(value, float):
                try:
                    return to_qvariant(format(value, self._format_spec))
                except (ValueError, TypeError):
                    # may happen if format = 'd' and value = NaN;
                    # see spyder-ide/spyder#4139.
                    return to_qvariant(format(value, DEFAULT_FORMAT))
            elif type(value) in [str, bytes]:
                # Don't perform any conversion on strings
                # because it leads to differences between
                # the data present in the dataframe and
                # what is shown by Spyder
                return value
            else:
                try:
                    return to_qvariant(str(value))
                except Exception:
                    self.display_error_idxs.append(index)
                    return u'Display Error!'
        elif role == Qt.BackgroundColorRole:
            return to_qvariant(self.get_bgcolor(index))
        elif role == Qt.FontRole:
            return self.get_font(SpyderFontType.MonospaceInterface)
        elif role == Qt.ToolTipRole:
            if index in self.display_error_idxs:
                return _("It is not possible to display this value because\n"
                         "an error occurred while trying to do it")
        return to_qvariant()

    def recalculate_index(self):
        """Recalcuate index information."""
        self.df_index_list = self.df.index.tolist()
        self.df_columns_list = self.df.columns.tolist()
        self.total_rows = self.df.shape[0]

        # Necessary to set rows_loaded because rowCount() method
        self.rows_loaded = self.df.shape[0]

        # Necessary to set cols_loaded because of columnCount() method
        self.cols_loaded = self.df.shape[1]
        self.total_cols = self.df.shape[1]

    def sort(self, column, order=Qt.AscendingOrder):
        """Overriding sort method"""
        if self.complex_intran is not None:
            if self.complex_intran.any(axis=0).iloc[column]:
                QMessageBox.critical(self.dialog, "Error",
                                     "TypeError error: no ordering "
                                     "relation is defined for complex numbers")
                return False
        try:
            ascending = order == Qt.AscendingOrder
            if column >= 0:
                try:
                    self.df.sort_values(by=self.df.columns[column],
                                        ascending=ascending, inplace=True,
                                        kind='mergesort')
                except AttributeError:
                    # for pandas version < 0.17
                    self.df.sort(columns=self.df.columns[column],
                                 ascending=ascending, inplace=True,
                                 kind='mergesort')
                except ValueError as e:
                    # Not possible to sort on duplicate columns
                    # See spyder-ide/spyder#5225.
                    QMessageBox.critical(self.dialog, "Error",
                                         "ValueError: %s" % str(e))
                except SystemError as e:
                    # Not possible to sort on category dtypes
                    # See spyder-ide/spyder#5361.
                    QMessageBox.critical(self.dialog, "Error",
                                         "SystemError: %s" % str(e))
            else:
                # Update index list
                self.recalculate_index()
                # To sort by index
                self.df.sort_index(inplace=True, ascending=ascending)
        except TypeError as e:
            QMessageBox.critical(self.dialog, "Error",
                                 "TypeError error: %s" % str(e))
            return False

        self.reset()
        return True

    def flags(self, index):
        """Set flags"""
        result = super().flags(index)
        if not self.readonly:
            result |= Qt.ItemFlag.ItemIsEditable
        return result

    def setData(self, index, value, role=Qt.EditRole):
        """Cell content change"""
        column = index.column()
        row = index.row()

        if index in self.display_error_idxs:
            return False

        val = from_qvariant(value, str)
        current_value = self.get_value(row, column)
        if isinstance(current_value, (bool, np.bool_)):
            val = bool_false_check(val)
        supported_types = (bool, np.bool_) + REAL_NUMBER_TYPES

        if (
            isinstance(current_value, supported_types)
            or isinstance(current_value, str)
        ):
            try:
                self.df.iloc[row, column] = current_value.__class__(val)
            except (ValueError, OverflowError) as e:
                QMessageBox.critical(self.dialog, "Error",
                                     str(type(e).__name__) + ": " + str(e))
                return False
        else:
            QMessageBox.critical(self.dialog, "Error",
                                 "Editing dtype {0!s} not yet supported."
                                 .format(type(current_value).__name__))
            return False

        self.max_min_col_update()
        self.dataChanged.emit(index, index)
        return True

    def get_data(self):
        """Return data"""
        return self.df

    def rowCount(self, index=QModelIndex()):
        """DataFrame row number"""
        # Avoid a "Qt exception in virtual methods" generated in our
        # tests on Windows/Python 3.7
        # See spyder-ide/spyder#8910.
        try:
            if self.total_rows <= self.rows_loaded:
                return self.total_rows
            else:
                return self.rows_loaded
        except AttributeError:
            return 0

    def fetch_more(self, rows=False, columns=False):
        """Get more columns and/or rows."""
        if rows and self.total_rows > self.rows_loaded:
            reminder = self.total_rows - self.rows_loaded
            items_to_fetch = min(reminder, ROWS_TO_LOAD)
            self.beginInsertRows(QModelIndex(), self.rows_loaded,
                                 self.rows_loaded + items_to_fetch - 1)
            self.rows_loaded += items_to_fetch
            self.endInsertRows()
        if columns and self.total_cols > self.cols_loaded:
            reminder = self.total_cols - self.cols_loaded
            items_to_fetch = min(reminder, COLS_TO_LOAD)
            self.beginInsertColumns(QModelIndex(), self.cols_loaded,
                                    self.cols_loaded + items_to_fetch - 1)
            self.cols_loaded += items_to_fetch
            self.endInsertColumns()

    def columnCount(self, index=QModelIndex()):
        """DataFrame column number"""
        # Avoid a "Qt exception in virtual methods" generated in our
        # tests on Windows/Python 3.7
        # See spyder-ide/spyder#8910.
        try:
            # This is done to implement series
            if len(self.df.shape) == 1:
                return 2
            elif self.total_cols <= self.cols_loaded:
                return self.total_cols
            else:
                return self.cols_loaded
        except AttributeError:
            return 0

    def reset(self):
        self.beginResetModel()
        self.endResetModel()


class DataFrameView(QTableView, SpyderWidgetMixin):
    """
    View displaying a dataframe in the dataframe editor

    This is a view (in the sense of the Qt model/view architecture) that is
    used in the dataframe editor to display a dataframe. It only shows the
    data but not the index (row headings) or header (column names).

    Parameters
    ----------
    parent : Optional[QWidget]
        The parent widget.
    model : DataFrameModel
        Model encapsulating the displayed dataframe.
    header : QHeaderView
        The header (column names) of the view.
    hscroll : QScrollBar
        The horizontal scroll bar.
    vscroll : QScrollBar
        The vertical scroll bar.
    namespacebrowser : Optional[NamespaceBrowser], optional
        The namespace browser that opened the editor containing this view.
        The default is None.
    data_function : Optional[Callable[[], Any]], optional
        A function which returns the new data frame when the user clicks on
        the Refresh button. The default is None.
    readonly : bool, optional
        If True, then the user can not edit the dataframe. The default is
        False.

    Signals
    -------
    sig_sort_by_column(): Raised after more columns are fetched.
    sig_fetch_more_rows(): Raised after more rows are fetched.
    """
    sig_sort_by_column = Signal()
    sig_fetch_more_columns = Signal()
    sig_fetch_more_rows = Signal()

    CONF_SECTION = 'variable_explorer'

    def __init__(
        self,
        parent: Optional[QWidget],
        model: DataFrameModel,
        header: QHeaderView,
        hscroll: QScrollBar,
        vscroll: QScrollBar,
        namespacebrowser: Optional[NamespaceBrowser] = None,
        data_function: Optional[Callable[[], Any]] = None,
        readonly: bool = False
    ):
        QTableView.__init__(self, parent)

        self.namespacebrowser = namespacebrowser
        self.data_function = data_function
        self.readonly = readonly

        self.menu = None
        self.menu_header_h = None
        self.empty_ws_menu = None
        self.copy_action = None
        self.edit_action = None
        self.edit_header_action = None
        self.insert_action_above = None
        self.insert_action_below = None
        self.insert_action_after = None
        self.insert_action_before = None
        self.remove_row_action = None
        self.remove_col_action = None
        self.duplicate_row_action = None
        self.duplicate_col_action = None
        self.resize_action = None
        self.resize_columns_action = None
        self.histogram_action = None

        self.menu = self.setup_menu()
        self.menu_header_h = self.setup_menu_header()
        self.register_shortcut_for_widget(name='copy', triggered=self.copy)

        self.setModel(model)
        self.setHorizontalScrollBar(hscroll)
        self.setVerticalScrollBar(vscroll)
        self.setHorizontalScrollMode(QTableView.ScrollPerPixel)
        self.setVerticalScrollMode(QTableView.ScrollPerPixel)

        self.sort_old = [None]
        self.header_class = header
        self.header_class.setContextMenuPolicy(Qt.CustomContextMenu)
        self.header_class.customContextMenuRequested.connect(
            self.show_header_menu)
        self.header_class.sectionClicked.connect(self.sortByColumn)
        self.horizontalScrollBar().valueChanged.connect(
            self._load_more_columns)
        self.verticalScrollBar().valueChanged.connect(self._load_more_rows)

    def _load_more_columns(self, value):
        """Load more columns to display."""
        # Needed to avoid a NameError while fetching data when closing
        # See spyder-ide/spyder#12034.
        try:
            self.load_more_data(value, columns=True)
        except NameError:
            pass

    def _load_more_rows(self, value):
        """Load more rows to display."""
        # Needed to avoid a NameError while fetching data when closing
        # See spyder-ide/spyder#12034.
        try:
            self.load_more_data(value, rows=True)
        except NameError:
            pass

    def load_more_data(self, value, rows=False, columns=False):
        """Load more rows and columns to display."""
        try:
            if rows and value == self.verticalScrollBar().maximum():
                self.model().fetch_more(rows=rows)
                self.sig_fetch_more_rows.emit()
            if columns and value == self.horizontalScrollBar().maximum():
                self.model().fetch_more(columns=columns)
                self.sig_fetch_more_columns.emit()

        except NameError:
            # Needed to handle a NameError while fetching data when closing
            # See spyder-ide/spyder#7880.
            pass

    def setModel(self, model: DataFrameModel) -> None:
        """
        Set the model for the view to present.

        This overrides the function in QTableView so that we can enable or
        disable actions when appropriate if the selection changes.
        """
        super().setModel(model)
        self.selectionModel().selectionChanged.connect(self.refresh_menu)
        self.refresh_menu()

    def sortByColumn(self, index):
        """Implement a column sort."""
        if self.sort_old == [None]:
            self.header_class.setSortIndicatorShown(True)
        sort_order = self.header_class.sortIndicatorOrder()
        if not self.model().sort(index, sort_order):
            if len(self.sort_old) != 2:
                self.header_class.setSortIndicatorShown(False)
            else:
                self.header_class.setSortIndicator(self.sort_old[0],
                                                   self.sort_old[1])
            return
        self.sort_old = [index, self.header_class.sortIndicatorOrder()]
        self.sig_sort_by_column.emit()

    def show_header_menu(self, pos):
        """Show edition menu for header."""
        global_pos = self.mapToGlobal(pos)
        index = self.indexAt(pos)
        self.header_class.setCurrentIndex(index)
        self.menu_header_h.popup(global_pos)

    def contextMenuEvent(self, event):
        """Reimplement Qt method."""
        self.menu.popup(event.globalPos())
        event.accept()

    def setup_menu_header(self):
        """Setup context header menu."""
        edit_header_action = self.create_action(
            name=DataframeEditorActions.EditHeader,
            text=_("Edit"),
            icon=ima.icon('edit'),
            triggered=self.edit_header_item,
            register_action=False
        )
        edit_header_action.setEnabled(not self.readonly)
        menu = self.create_menu(DataframeEditorMenus.Header, register=False)
        self.add_item_to_menu(edit_header_action, menu)
        return menu

    def refresh_menu(self):
        """Refresh context menu"""
        index = self.currentIndex()

        # Enable/disable edit actions
        condition_edit = (
            index.isValid()
            and len(self.selectedIndexes()) == 1
            and not self.readonly
        )

        for action in [self.edit_action, self.insert_action_above,
                       self.insert_action_below, self.insert_action_after,
                       self.insert_action_before, self.duplicate_row_action,
                       self.duplicate_col_action]:
            action.setEnabled(condition_edit)

        # Enable/disable actions for remove col/row, copy and plot
        condition_copy_remove = (
            index.isValid()
            and len(self.selectedIndexes()) > 0
            and not self.readonly
        )

        for action in [self.copy_action, self.remove_row_action,
                       self.remove_col_action, self.histogram_action]:
            action.setEnabled(condition_copy_remove)

        # Enable/disable action for plot
        condition_plot = (index.isValid() and len(self.selectedIndexes()) > 0)
        self.histogram_action.setEnabled(condition_plot)

    def setup_menu(self):
        """Setup context menu."""
        # ---- Create actions

        self.resize_action = self.create_action(
            name=DataframeEditorActions.ResizeRows,
            text=_("Resize rows to contents"),
            icon=ima.icon('collapse_row'),
            triggered=lambda: self.resize_to_contents(rows=True),
            register_action=False
        )
        self.resize_columns_action = self.create_action(
            name=DataframeEditorActions.ResizeColumns,
            text=_("Resize columns to contents"),
            icon=ima.icon('collapse_column'),
            triggered=self.resize_to_contents,
            register_action=False
        )
        self.edit_action = self.create_action(
            name=DataframeEditorActions.Edit,
            text=_("Edit"),
            icon=ima.icon('edit'),
            triggered=self.edit_item,
            register_action=False
        )
        self.insert_action_above = self.create_action(
            name=DataframeEditorActions.InsertAbove,
            text=_("Insert above"),
            icon=ima.icon('insert_above'),
            triggered=lambda: self.insert_item(axis=1, before_above=True),
            register_action=False
        )
        self.insert_action_below = self.create_action(
            name=DataframeEditorActions.InsertBelow,
            text=_("Insert below"),
            icon=ima.icon('insert_below'),
            triggered=lambda: self.insert_item(axis=1, before_above=False),
            register_action=False
        )
        self.insert_action_before = self.create_action(
            name=DataframeEditorActions.InsertBefore,
            text=_("Insert before"),
            icon=ima.icon('insert_before'),
            triggered=lambda: self.insert_item(axis=0, before_above=True),
            register_action=False
        )
        self.insert_action_after = self.create_action(
            name=DataframeEditorActions.InsertAfter,
            text=_("Insert after"),
            icon=ima.icon('insert_after'),
            triggered=lambda: self.insert_item(axis=0, before_above=False),
            register_action=False
        )
        self.remove_row_action = self.create_action(
            name=DataframeEditorActions.RemoveRow,
            text=_("Remove row"),
            icon=ima.icon('delete_row'),
            triggered=self.remove_item,
            register_action=False
        )
        self.remove_col_action = self.create_action(
            name=DataframeEditorActions.RemoveColumn,
            text=_("Remove column"),
            icon=ima.icon('delete_column'),
            triggered=lambda: self.remove_item(axis=1),
            register_action=False
        )
        self.duplicate_row_action = self.create_action(
            name=DataframeEditorActions.DuplicateRow,
            text=_("Duplicate row"),
            icon=ima.icon('duplicate_row'),
            triggered=lambda: self.duplicate_row_col(dup_row=True),
            register_action=False
        )
        self.duplicate_col_action = self.create_action(
            name=DataframeEditorActions.DuplicateColumn,
            text=_("Duplicate column"),
            icon=ima.icon('duplicate_column'),
            triggered=lambda: self.duplicate_row_col(dup_row=False),
            register_action=False
        )
        self.copy_action = self.create_action(
            name=DataframeEditorActions.Copy,
            text=_('Copy'),
            icon=ima.icon('editcopy'),
            triggered=self.copy,
            register_action=False
        )
        self.copy_action.setShortcut(keybinding('Copy'))
        self.copy_action.setShortcutContext(Qt.WidgetShortcut)
        self.histogram_action = self.create_action(
            name=DataframeEditorActions.Histogram,
            text=_("Histogram"),
            tip=_("Plot a histogram of the selected columns"),
            icon=ima.icon('hist'),
            triggered=self.plot_hist,
            register_action=False
        )

        # ---- Create context menu and fill it

        menu = self.create_menu(DataframeEditorMenus.Context, register=False)
        for action in [self.copy_action, self.edit_action]:
            self.add_item_to_menu(
                action,
                menu,
                section=DataframeEditorContextMenuSections.Edit
            )
        for action in [self.insert_action_above, self.insert_action_below,
                       self.duplicate_row_action, self.remove_row_action]:
            self.add_item_to_menu(
                action,
                menu,
                section=DataframeEditorContextMenuSections.Row
            )
        for action in [self.insert_action_before, self.insert_action_after,
                       self.duplicate_col_action, self.remove_col_action]:
            self.add_item_to_menu(
                action,
                menu,
                section=DataframeEditorContextMenuSections.Column
            )

        return menu

    @Slot()
    def copy(self):
        """Copy text to clipboard"""
        if not self.selectedIndexes():
            return
        (row_min, row_max,
         col_min, col_max) = get_idx_rect(self.selectedIndexes())
        # Copy index and header too (equal True).
        # See spyder-ide/spyder#11096
        index = header = True
        df = self.model().df
        obj = df.iloc[slice(row_min, row_max + 1),
                      slice(col_min, col_max + 1)]
        output = io.StringIO()
        try:
            obj.to_csv(output, sep='\t', index=index, header=header)
        except UnicodeEncodeError:
            # Needed to handle encoding errors in Python 2
            # See spyder-ide/spyder#4833
            QMessageBox.critical(
                self,
                _("Error"),
                _("Text can't be copied."))
        contents = output.getvalue()
        output.close()
        clipboard = QApplication.clipboard()
        clipboard.setText(contents)

    def resize_to_contents(self, rows=False):
        """Resize rows or cols to its contents."""
        if isinstance(self.parent(), DataFrameEditor):
            if rows:
                self.resizeRowsToContents()
                self.parent().table_index.resizeRowsToContents()
            else:
                self.parent().table_index.resizeColumnsToContents()
                self.parent().resize_to_contents()

    def edit_header_item(self):
        """Edit header item"""
        pos = self.header_class.currentIndex()
        index = self.header_class.logicalIndex(pos.column())
        if index >= 0:
            model_index = self.header_class.model().index(0, index)
            index_number_rows = 1

            if type(self.model().df.columns[0]) is tuple:
                index_number_rows = len(self.model().df.columns[0])

            if index_number_rows > 1:
                dialog = QInputDialog()
                dialog.setWindowTitle("Enter the values")
                label = QLabel("Enter the values:")
                dialog.show()
                dialog.findChild(QLineEdit).hide()
                dialog.findChild(QLabel).hide()
                lines = []
                for row in range(index_number_rows):
                    line = QLineEdit(text=self.model().df.columns[index][row])
                    dialog.layout().insertWidget(row, line)
                    lines.append(line)
                dialog.layout().insertWidget(0, label)
                dialog.hide()
                confirmation = dialog.exec_() == QDialog.Accepted
                if confirmation:
                    value = tuple(line.text() for line in lines)
            else:
                value, confirmation = QInputDialog.getText(
                    self,
                    _("Enter a value"),
                    _("Enter a value"),
                    QLineEdit.Normal,
                    ""
                )

            if confirmation:
                if value not in self.model().df.columns.tolist():
                    if type(value) is tuple:
                        n_cols = len(self.model().df.columns)
                        cols = self.model().df.columns
                        names = cols.names
                        cols = (
                            self.model().df.columns.tolist()[0:index]
                            + [value]
                            + self.model().df.columns.tolist()[index+1:n_cols]
                        )
                        self.model().df.columns = (
                            pd.MultiIndex.from_tuples(cols, names=names)
                        )
                    else:
                        self.header_class.model().setData(
                            model_index,
                            value,
                            Qt.EditRole
                        )

                    self.parent()._reload()
                    self.model().dataChanged.emit(pos, pos)
                else:
                    QMessageBox.warning(
                        self.model().dialog,
                        _("Warning: Duplicate column"),
                        _('Column with name "{}" already exists!').format(
                            value)
                    )

    def edit_item(self):
        """Edit item"""
        index = self.currentIndex()
        if not index.isValid():
            return

        self.edit(index)

    def insert_item(self, axis=0, before_above=False):
        """Insert row or column."""
        current_index = self.currentIndex()
        if not current_index.isValid():
            return False

        column = current_index.column()
        row = current_index.row()
        step = 0
        df = self.model().df

        if not before_above:
            step = 1

        if axis == 0:
            # insert column
            module = df.iat[row, column].__class__.__module__

            if module == 'builtins':
                # Evaluate character '' (empty) to initialize the column as a
                # neutral data type
                eval_type = df.iat[row, column].__class__.__name__ + '('')'
            else:
                # Necessary because of import numpy as np
                if module == 'numpy':
                    module = 'np'
                eval_type = (
                    module
                    + '.'
                    + df.iat[row, column].__class__.__name__
                    + '('')'
                )

            indexes = df.axes[1].tolist()
            new_name = 'new_col'
            if type(indexes[column]) is not str:
                new_name = indexes[column]

            if new_name in indexes:
                if type(new_name) is tuple:
                    tuple_idx = []
                    new_tuple = []

                    for idx in indexes:
                        tuple_idx = tuple_idx + list(idx)

                    for idx in range(len(new_name)):
                        new_tuple.append(
                            self.next_index_name(tuple_idx, new_name[idx])
                        )

                    new_name = tuple(new_tuple)
                else:
                    new_name = self.next_index_name(indexes, new_name)

            item_value = eval(eval_type)
            if isinstance(item_value, tuple) and item_value == ():
                item_value = ('')

            df.insert(
                loc=column + step,
                column=new_name,
                value=item_value,
                allow_duplicates=True
            )

            self.model().max_min_col_update()
            if before_above:
                column = column + 1

        if axis == 1:
            # insert row
            indexes = df.axes[0].tolist()
            new_name = 'new_row'
            if type(indexes[row]) is not str:
                new_name = indexes[row]
            if new_name in indexes:
                new_name = self.next_index_name(indexes, new_name)

            # Slice the upper half of the dataframe
            df1 = df[0:row + step]

            # Store the result of lower half of the dataframe
            df2 = df[row + step:]

            # Insert the row in the upper half dataframe
            new_row = df.iloc[[row]]
            if parse(pd.__version__) >= parse("3.0.0"):
                new_row.index = [new_name]
            else:
                new_row.axes[0].values[0] = new_name

            for col in range(len(new_row.columns)):
                module = new_row.iat[0, col].__class__.__module__
                if module == 'builtins':
                    # Evaluate character '' (empty) to initialyze the column as
                    # a neutral data type
                    eval_type = new_row.iat[0, col].__class__.__name__ + '('')'
                else:
                    # Necessary because of import numpy as np
                    if module == 'numpy':
                        module = 'np'
                    eval_type = (
                        module
                        + '.'
                        + new_row.iat[0, col].__class__.__name__
                        + '('')'
                    )

                new_row.iat[0, col] = eval(eval_type)

            self.model().df = pd.concat([df1, new_row, df2])
            if before_above:
                row = row + 1

        self.parent()._reload()
        self.model().dataChanged.emit(current_index, current_index)
        self.setCurrentIndex(self.model().index(row, column))

    def duplicate_row_col(self, dup_row=False):
        """Duplicate row or column."""
        current_index = self.currentIndex()
        if not current_index.isValid():
            return False

        column = current_index.column()
        row = current_index.row()
        df = self.model().df

        if dup_row:
            # Slice the upper half of the dataframe
            df1 = self.model().df[0:row]

            # Store the result of lower half of the dataframe
            df2 = self.model().df[row:]

            # Insert the row in the upper half dataframe
            new_row = self.model().df.iloc[[row]]
            label = new_row.axes[0].values[0]
            indexes = self.model().df.axes[0].tolist()
            indexes.remove(label)

            new_name = self.next_index_name(indexes, label)
            if parse(pd.__version__) >= parse("3.0.0"):
                new_row.index = [new_name]
            else:
                new_row.axes[0].values[0] = new_name

            self.model().df = pd.concat([df1, new_row, df2])
            row = row + 1
        else:
            indexes = df.axes[1].tolist()
            label = indexes[column]
            indexes.remove(label)

            if type(label) is tuple:
                tuple_idx = []
                new_tuple = []

                for idx in indexes:
                    tuple_idx = tuple_idx + list(idx)

                for idx in range(len(label)):
                    new_tuple.append(
                        self.next_index_name(tuple_idx, label[idx])
                    )
                new_name = tuple(new_tuple)
            else:
                new_name = self.next_index_name(indexes, label)

            df.insert(loc=column+1, column=new_name, value='',
                      allow_duplicates=True)
            df[new_name] = df.iloc[:, column]
            self.model().max_min_col_update()

        self.parent()._reload()
        self.model().dataChanged.emit(current_index, current_index)
        self.setCurrentIndex(self.model().index(row, column))

    def next_index_name(self, indexes, label):
        """
        Calculate and generate next index_name for a duplicate column/row
        rol/col_copy(ind).
        """
        ind = -1
        name = ''
        acceptable_types = (
            [str, float, int, complex, bool]
            + list(REAL_NUMBER_TYPES)
            + list(COMPLEX_NUMBER_TYPES)
        )

        if type(label) not in acceptable_types:
            # Case receiving a different type of acceptable_type,
            # treat as string
            label = str(label)

        if type(label) is str:
            # Make all indexes strings to compare
            for i in range(len(indexes)):
                if type(indexes[i]) is not str:
                    indexes[i] = str(indexes[i])

            # Verify if find '_copy(' in the label
            if label.rfind('_copy(') == -1:
                # If not found, verify in other indexes
                name = label + '_copy('

                for n in indexes:
                    if n.rfind(name) == 0:
                        # label_copy( starts in first position
                        init_pos = len(name)
                        final_pos = len(n) - 1
                        curr_ind = n[init_pos:final_pos]
                        if (
                            curr_ind.isnumeric()
                            and n[final_pos:final_pos+1] == ')'
                        ):
                            if ind < int(curr_ind):
                                ind = int(curr_ind)
            else:
                # If 'copy_(' string is in label, verify if valid and check
                # next.
                init_pos = label.rfind('_copy(') + 6
                final_pos = len(label) - 1
                curr_ind = label[init_pos:final_pos]

                if curr_ind.isnumeric():
                    if label[final_pos:final_pos+1] == ')':
                        ind = int(curr_ind)
                        name = label[0:init_pos]

                        for n in indexes:
                            if n.rfind(name) == 0:
                                init_pos = len(name)
                                final_pos = len(n) - 1
                                curr_ind = n[init_pos:final_pos]
                                if (
                                    curr_ind.isnumeric()
                                    and n[final_pos:final_pos+1] == ')'
                                ):
                                    if ind < int(curr_ind):
                                        ind = int(curr_ind)
                    else:
                        # If not closed parenthesis, treat entire string as
                        # valid
                        name = label + '_copy('
                        for n in indexes:
                            if n.rfind(name) == 0:
                                init_pos = len(name)
                                final_pos = len(n) - 1
                                curr_ind = n[init_pos:final_pos]
                                if (
                                    curr_ind.isnumeric()
                                    and n[final_pos:final_pos+1] == ')'
                                ):
                                    if ind < int(curr_ind):
                                        ind = int(curr_ind)
                else:
                    # Found '_copy(not a number)', treat entire string as valid
                    # and check if exist other '_copy(Not number)*_copy(number)
                    name = label + '_copy('
                    for n in indexes:
                        if n.rfind(name) == 0:
                            init_pos = len(name)
                            final_pos = len(n) - 1
                            curr_ind = n[init_pos:final_pos]
                            if (
                                curr_ind.isnumeric()
                                and n[final_pos:final_pos+1] == ')'
                            ):
                                if ind < int(curr_ind):
                                    ind = int(curr_ind)

            ind = ind+1
            return name + str(ind) + ')'
        else:
            # Type is numeric: increment 1 and check if it is in list.
            label = label + 1

            while label in indexes:
                label = label + 1

            return label

    @Slot()
    def remove_item(self, force=False, axis=0):
        """Remove item."""
        indexes = self.selectedIndexes()
        index_label = []
        df = self.model().df
        if not indexes:
            return

        # Keep focus on the item before the deleted one
        focus_row = indexes[0].row()
        focus_col = indexes[0].column()
        if axis == 0 and focus_row > 0:
            focus_row = focus_row - 1
        if axis == 1 and focus_col > 0:
            focus_col = focus_col - 1

        for index in indexes:
            if not index.isValid():
                return
            else:
                if axis == 0:
                    row_label = df.axes[axis][index.row()]
                    if row_label not in index_label:
                        index_label.append(row_label)
                else:
                    column_label = df.axes[axis][index.column()]
                    if column_label not in index_label:
                        index_label.append(column_label)

        result = None
        if not force:
            if (
                not self.get_conf('show_remove_message_dataframe')
                or running_under_pytest()
            ):
                result = QMessageBox.Yes
            else:
                one = _("Do you want to remove the selected item?")
                more = _("Do you want to remove all selected items?")
                answer = MessageCheckBox(
                    icon=QMessageBox.Question, parent=self
                )
                answer.set_checkbox_text(_("Don't ask again."))
                answer.set_checked(False)
                answer.set_check_visible(True)
                answer.setText(one if len(indexes) == 1 else more)
                answer.setStandardButtons(QMessageBox.Yes | QMessageBox.No)

                result = answer.exec_()
                check = answer.is_checked()
                if check:
                    self.set_conf('show_remove_message_dataframe', False)

        if force or result == QMessageBox.Yes:
            for label in index_label:
                try:
                    df.drop(label, inplace=True, axis=axis)
                except TypeError as e:
                    QMessageBox.warning(
                        self.model().dialog,
                        _("Warning: It was not possible to remove this item!"),
                        _("ValueError: {} must be removed from index.").format(
                            str(e))
                    )
                    return False

            self.model().max_min_col_update()
            self.parent()._reload()
            index = QModelIndex()
            self.model().dataChanged.emit(index, index)
            self.setCurrentIndex(self.model().index(focus_row, focus_col))

    def plot_hist(self) -> None:
        """
        Plot histogram of selected columns
        """
        def plot_function(figure: Figure) -> None:
            ax = figure.subplots()
            model.df.hist(column=col_labels, ax=ax)

        cols = list(index.column() for index in self.selectedIndexes())
        cols = list(set(cols))  # Remove duplicates
        model = self.model()
        col_labels = [model.header(0, col) for col in cols]
        self.namespacebrowser.plot(plot_function)


class DataFrameHeaderModel(QAbstractTableModel, SpyderFontsMixin):
    """
    This class is the model for the header and index of the DataFrameEditor.

    Taken from gtabview project (Header4ExtModel).
    For more information please see:
    https://github.com/wavexx/gtabview/blob/master/gtabview/viewer.py
    """

    COLUMN_INDEX = -1  # Makes reference to the index of the table.

    def __init__(self, model, axis, use_monospace_font=False):
        """
        Header constructor.

        The 'model' is the QAbstractTableModel of the dataframe, the 'axis' is
        to acknowledge if is for the header (horizontal - 0) or for the
        index (vertical - 1) and the palette is the set of colors to use.
        """
        super().__init__()
        self.model = model
        self.axis = axis
        self.use_monospace_font = use_monospace_font

        self.total_rows = self.model.shape[0]
        self.total_cols = self.model.shape[1]
        self.cols_loaded = self.model.cols_loaded
        self.rows_loaded = self.model.rows_loaded

        if self.axis == 0:
            self.total_cols = self.model.shape[1]
            self._shape = (self.model.header_shape[0], self.model.shape[1])
        else:
            self.total_rows = self.model.shape[0]
            self._shape = (self.model.shape[0], self.model.header_shape[1])

    def rowCount(self, index=None):
        """Get number of rows in the header."""
        if self.axis == 0:
            return max(1, self._shape[0])
        else:
            if self.total_rows <= self.rows_loaded:
                return self.total_rows
            else:
                return self.rows_loaded

    def columnCount(self, index=QModelIndex()):
        """DataFrame column number"""
        if self.axis == 0:
            if self.total_cols <= self.cols_loaded:
                return self.total_cols
            else:
                return self.cols_loaded
        else:
            return max(1, self._shape[1])

    def fetch_more(self, rows=False, columns=False):
        """Get more columns or rows (based on axis)."""
        if self.axis == 1 and self.total_rows > self.rows_loaded:
            reminder = self.total_rows - self.rows_loaded
            items_to_fetch = min(reminder, ROWS_TO_LOAD)
            self.beginInsertRows(QModelIndex(), self.rows_loaded,
                                 self.rows_loaded + items_to_fetch - 1)
            self.rows_loaded += items_to_fetch
            self.endInsertRows()
        if self.axis == 0 and self.total_cols > self.cols_loaded:
            reminder = self.total_cols - self.cols_loaded
            items_to_fetch = min(reminder, COLS_TO_LOAD)
            self.beginInsertColumns(QModelIndex(), self.cols_loaded,
                                    self.cols_loaded + items_to_fetch - 1)
            self.cols_loaded += items_to_fetch
            self.endInsertColumns()

    def sort(self, column, order=Qt.AscendingOrder):
        """Overriding sort method."""
        ascending = order == Qt.AscendingOrder
        self.model.sort(self.COLUMN_INDEX, order=ascending)
        return True

    def headerData(self, section, orientation, role):
        """Get the information to put in the header."""
        if role == Qt.TextAlignmentRole:
            if orientation == Qt.Horizontal:
                return Qt.AlignCenter
            else:
                return int(Qt.AlignRight | Qt.AlignVCenter)
        if role != Qt.DisplayRole and role != Qt.ToolTipRole:
            return None
        if self.axis == 1 and self._shape[1] <= 1:
            return None
        orient_axis = 0 if orientation == Qt.Horizontal else 1
        if self.model.header_shape[orient_axis] > 1:
            header = section
        else:
            header = self.model.header(self.axis, section)

            # Don't perform any conversion on strings
            # because it leads to differences between
            # the data present in the dataframe and
            # what is shown by Spyder
            if not type(header) in [str, bytes]:
                header = str(header)

        return header

    def data(self, index, role):
        """
        Get the data for the header.

        This is used when a header has levels.
        """
        if (
            not index.isValid()
            or index.row() >= self._shape[0]
            or index.column() >= self._shape[1]
        ):
            return None

        row, col = (
            (index.row(), index.column()) if self.axis == 0
            else (index.column(), index.row())
        )

        if self.use_monospace_font and role == Qt.FontRole:
            return self.get_font(SpyderFontType.MonospaceInterface)

        if role != Qt.DisplayRole:
            return None

        if self.axis == 0 and self._shape[0] <= 1:
            return None

        header = self.model.header(self.axis, col, row)

        # Don't perform any conversion on strings
        # because it leads to differences between
        # the data present in the dataframe and
        # what is shown by Spyder
        if not type(header) in [str, bytes]:
            header = str(header)

        return header

    def flags(self, index):
        """Set flags"""
        return (QAbstractTableModel.flags(self, index) |
                Qt.ItemFlag.ItemIsEditable |
                Qt.ItemFlag.ItemIsEnabled |
                Qt.ItemFlag.ItemIsSelectable
        )

    def setData(self, index, value, role):
        """Cell content change"""
        df = self.model.df

        if role == Qt.EditRole:
            if self.axis == 1:
                old_value = df.index[index.row()]

                if value not in df.index.tolist():
                    if type(old_value) is tuple:
                        old_value_list = list(old_value)
                        rows = df.index
                        names = rows.names
                        old_value_list[index.column()] = value
                        rows = (
                            df.index.tolist()[0:index.row()]
                            + [tuple(old_value_list)]
                            + df.index.tolist()[index.row()+1:]
                        )
                        df.index = pd.MultiIndex.from_tuples(rows, names=names)
                    else:
                        try:
                            df.rename(index={old_value: value}, inplace=True,
                                      errors='raise')
                        except TypeError as e:
                            QMessageBox.warning(
                                self.model().dialog,
                                _("Warning: It was not possible to remove "
                                  "this index!"),
                                _("ValueError: {} must be removed from "
                                  "index.").format(str(e))
                            )
                            return False
                else:
                    QMessageBox.warning(
                        self.model().dialog,
                        _("Warning: Duplicate index!"),
                        _('Row with name "{}" already exists!').format(value)
                    )
                    return False

                self.model.dialog._reload()
                self.model.dataChanged.emit(index, index)
                return True

            if self.axis == 0:
                old_value = df.columns[index.column()]

                try:
                    df.rename(columns={old_value: value}, inplace=True,
                              errors='raise')
                except Exception:
                    return False

                return True

            return True

        return False


class DataFrameLevelModel(QAbstractTableModel, SpyderFontsMixin):
    """
    Data Frame level class.

    This class is used to represent index levels in the DataFrameEditor. When
    using MultiIndex, this model creates labels for the index/header as Index i
    for each section in the index/header

    Based on the gtabview project (Level4ExtModel).
    For more information please see:
    https://github.com/wavexx/gtabview/blob/master/gtabview/viewer.py
    """

    def __init__(self, model):
        super().__init__()
        self.model = model
        self._background = QColor(SpyderPalette.COLOR_BACKGROUND_2)

    def rowCount(self, index=None):
        """Get number of rows (number of levels for the header)."""
        return max(1, self.model.header_shape[0])

    def columnCount(self, index=None):
        """Get the number of columns (number of levels for the index)."""
        return max(1, self.model.header_shape[1])

    def headerData(self, section, orientation, role):
        """
        Get the text to put in the header of the levels of the indexes.

        By default it returns 'Index i', where i is the section in the index
        """
        if role == Qt.TextAlignmentRole:
            if orientation == Qt.Horizontal:
                return Qt.AlignCenter
            else:
                return int(Qt.AlignRight | Qt.AlignVCenter)
        if role != Qt.DisplayRole and role != Qt.ToolTipRole:
            return None
        if self.model.header_shape[0] <= 1 and orientation == Qt.Horizontal:
            if self.model.name(1, section):
                return self.model.name(1, section)
            return _('Index')
        elif self.model.header_shape[0] <= 1:
            return None
        elif self.model.header_shape[1] <= 1 and orientation == Qt.Vertical:
            return None
        return _('Index') + ' ' + str(section)

    def data(self, index, role):
        """Get the information of the levels."""
        if not index.isValid():
            return None
        if role == Qt.FontRole:
            return self.get_font(SpyderFontType.Interface)
        label = ''
        if index.column() == self.model.header_shape[1] - 1:
            label = str(self.model.name(0, index.row()))
        elif index.row() == self.model.header_shape[0] - 1:
            label = str(self.model.name(1, index.column()))
        if role == Qt.DisplayRole and label:
            return label
        elif role == Qt.BackgroundRole:
            return self._background
        elif role == Qt.BackgroundRole:
            return self._palette.window()
        return None


class EmptyDataFrame:
    shape = (0, 0)


class DataFrameEditor(BaseDialog, SpyderWidgetMixin):
    """
    Dialog for displaying and editing DataFrame and related objects.

    The main portion of the dialog is a table displaying the data. The column
    names are displayed above the table and the row index is displayed to the
    left of the table. Above this all is a toolbar and below it all are buttons
    for closing the dialog and saving the data.

    Based on the gtabview project (ExtTableView).
    For more information please see:
    https://github.com/wavexx/gtabview/blob/master/gtabview/viewer.py

    Parameters
    ----------
    parent : Optional[QWidget]
        The parent widget.
    namespacebrowser : Optional[NamespaceBrowser], optional
        The namespace browser that opened the editor containing this view.
        The default is None.
    data_function : Optional[Callable[[], Any]], optional
        A function which returns the new data frame when the user clicks on
        the Refresh button. The default is None.
    readonly : bool, optional
        If True, then the user can not edit the dataframe. The default is
        False.
    """
    CONF_SECTION = 'variable_explorer'

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        namespacebrowser: Optional[NamespaceBrowser] = None,
        data_function: Optional[Callable[[], Any]] = None,
        readonly: bool = False
    ):
        super().__init__(parent)
        self.namespacebrowser = namespacebrowser
        self.data_function = data_function
        self.readonly = readonly

        self.refresh_action = self.create_action(
            name=DataframeEditorActions.Refresh,
            text=_('Refresh'),
            icon=ima.icon('refresh'),
            tip=_('Refresh editor with current value of variable in console'),
            triggered=self.refresh_editor,
            register_action=False
        )
        self.refresh_action.setEnabled(self.data_function is not None)
        self.preferences_action = self.create_action(
            name=DataframeEditorActions.Preferences,
            icon=self.create_icon('configure'),
            text=_('Display options ...'),
            triggered=self.show_preferences_dialog,
            register_action=False
        )
        self.close_action = self.create_action(
            name=DataframeEditorActions.Close,
            icon=self.create_icon('close_pane'),
            text=_('Close'),
            triggered=self.reject,
            shortcut=self.get_shortcut(DataframeEditorActions.Close),
            register_action=False,
            register_shortcut=True
        )
        self.register_shortcut_for_widget(name='close', triggered=self.reject)

        # Destroying the C++ object right after closing the dialog box,
        # otherwise it may be garbage-collected in another QThread
        # (e.g. the editor's analysis thread in Spyder), thus leading to
        # a segmentation fault on UNIX or an application crash on Windows
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.is_series = False
        self.layout = None
        self.glayout = None
        self.menu_header_v = None
        self.dataTable = None
        self.resizeToHeader = False

    def setup_and_check(self, data, title='') -> bool:
        """
        Setup editor.

        It returns False if data is not supported, True otherwise. Supported
        types for data are DataFrame, Series and Index.
        """
        if title:
            title = str(title) + " - %s" % data.__class__.__name__
        else:
            title = _("%s editor") % data.__class__.__name__

        self.setup_ui(title)
        return self.set_data_and_check(data)

    def setup_ui(self, title: str) -> None:
        """
        Create user interface.
        """
        # ---- Toolbar (to be filled later)

        self.toolbar = self.create_toolbar(
            DataframeEditorWidgets.Toolbar,
            register=False
        )

        # ---- Grid layout with tables and scrollbars showing data frame

        self.glayout = QGridLayout()
        self.glayout.setSpacing(0)
        self.glayout.setContentsMargins(0, 0, 0, 0)

        self.hscroll = QScrollBar(Qt.Horizontal)
        self.vscroll = QScrollBar(Qt.Vertical)

        # Create the view for the level
        self.create_table_level()

        # Create the view for the horizontal header
        self.create_table_header()

        # Create the view for the vertical index
        self.create_table_index()

        # Create menu to allow edit index
        self.menu_header_v = self.setup_menu_header(self.table_index)

        # Create the model and view of the data
        empty_data = EmptyDataFrame()
        self.dataModel = DataFrameModel(
            empty_data,
            parent=self,
            readonly=self.readonly
        )
        self.dataModel.dataChanged.connect(self.save_and_close_enable)
        self.create_data_table()

        self.glayout.addWidget(self.hscroll, 2, 0, 1, 2)
        self.glayout.addWidget(self.vscroll, 0, 2, 2, 1)

        # autosize columns on-demand
        self._autosized_cols = set()

        # Set limit time to calculate column sizeHint to 300ms,
        # See spyder-ide/spyder#11060
        self._max_autosize_ms = 300
        self.dataTable.installEventFilter(self)

        avg_width = self.fontMetrics().averageCharWidth()
        self.min_trunc = avg_width * 12  # Minimum size for columns
        self.max_width = avg_width * 64  # Maximum size for columns

        # ---- Buttons at bottom

        btn_layout = QHBoxLayout()
        btn_layout.addStretch()

        if self.readonly:
            self.btn_save_and_close = None
        else:
            self.btn_save_and_close = QPushButton(_('Save and Close'))
            self.btn_save_and_close.clicked.connect(self.accept)
            btn_layout.addWidget(self.btn_save_and_close)

        self.btn_close = QPushButton(_('Close'))
        self.btn_close.setAutoDefault(True)
        self.btn_close.setDefault(True)
        self.btn_close.clicked.connect(self.reject)
        btn_layout.addWidget(self.btn_close)

        # ---- Final layout

        self.layout = QVBoxLayout()
        self.layout.addWidget(self.toolbar)

        # Remove vertical space between toolbar and data frame
        style = self.style()
        default_spacing = style.pixelMetric(QStyle.PM_LayoutVerticalSpacing)
        self.layout.addSpacing(-default_spacing)

        self.layout.addLayout(self.glayout)
        self.layout.addSpacing((-1 if MAC else 2) * AppStyle.MarginSize)
        self.layout.addLayout(btn_layout)
        self.setLayout(self.layout)

        self.setWindowTitle(title)

        # Make the dialog act as a window
        self.setWindowFlags(Qt.Window)

    def set_data_and_check(self, data) -> bool:
        """
        Checks whether data is suitable and display it in the editor.

        This method returns False if data is not supported.
        """
        if not isinstance(data, (pd.DataFrame, pd.Series, pd.Index)):
            return False

        self._selection_rec = False
        self._model = None

        if isinstance(data, pd.Series):
            self.is_series = True
            data = data.to_frame()
        elif isinstance(data, pd.Index):
            data = pd.DataFrame(data)

        # Create the model and view of the data
        self.dataModel = DataFrameModel(
            data,
            parent=self,
            readonly=self.readonly
        )
        self.dataModel.dataChanged.connect(self.save_and_close_enable)
        self.dataTable.setModel(self.dataModel)

        # autosize columns on-demand
        self._autosized_cols = set()

        # Set limit time to calculate column sizeHint to 300ms,
        # See spyder-ide/spyder#11060
        self._max_autosize_ms = 300
        self.dataTable.installEventFilter(self)

        self.setModel(self.dataModel)
        self.resizeColumnsToContents()

        if self.btn_save_and_close:
            self.btn_save_and_close.setDisabled(True)
        self.dataModel.set_format_spec(self.get_conf('dataframe_format'))

        if self.table_header.rowHeight(0) == 0:
            self.table_header.setRowHeight(0, self.table_header.height())

        stretcher = self.create_stretcher(
            DataframeEditorWidgets.ToolbarStretcher
        )

        options_menu = self.create_menu(
            DataframeEditorMenus.Options,
            register=False
        )
        for item in [self.preferences_action, self.close_action]:
            self.add_item_to_menu(item, options_menu)

        options_button = self.create_toolbutton(
            name=DataframeEditorWidgets.OptionsToolButton,
            text=_('Options'),
            icon=ima.icon('tooloptions'),
            register=False
        )
        options_button.setPopupMode(QToolButton.InstantPopup)
        options_button.setMenu(options_menu)

        self.toolbar.clear()
        self.toolbar._section_items.clear()
        self.toolbar._item_map.clear()

        for item in [
            self.dataTable.insert_action_above,
            self.dataTable.insert_action_below,
            self.dataTable.duplicate_row_action,
            self.dataTable.remove_row_action
        ]:
            self.add_item_to_toolbar(
                item,
                self.toolbar,
                section=DataframeEditorToolbarSections.Row
            )

        for item in [
            self.dataTable.insert_action_before,
            self.dataTable.insert_action_after,
            self.dataTable.duplicate_col_action,
            self.dataTable.remove_col_action,
            stretcher,
            self.dataTable.histogram_action,
            self.dataTable.resize_action,
            self.dataTable.resize_columns_action,
            self.refresh_action,
            options_button
        ]:
            self.add_item_to_toolbar(
                item,
                self.toolbar,
                section=DataframeEditorToolbarSections.ColumnAndRest
            )

        self.toolbar.render()

        return True

    @Slot(QModelIndex, QModelIndex)
    def save_and_close_enable(self, top_left, bottom_right):
        """Handle the data change event to enable the save and close button."""
        if self.btn_save_and_close:
            self.btn_save_and_close.setEnabled(True)
            self.btn_save_and_close.setAutoDefault(True)
            self.btn_save_and_close.setDefault(True)

    def setup_menu_header(self, header):
        """Setup context header menu."""
        edit_header_action = self.create_action(
            name=DataframeEditorActions.EditIndex,
            text=_("Edit"),
            icon=ima.icon('edit'),
            triggered=lambda: self.edit_header_item(header=header),
            register_action=False
        )
        edit_header_action.setEnabled(not self.readonly)
        menu = self.create_menu(DataframeEditorMenus.Index, register=False)
        self.add_item_to_menu(edit_header_action, menu)
        return menu

    @Slot()
    def edit_header_item(self, header=None):
        """Edit item"""

        index = header.currentIndex()
        header.setUpdatesEnabled(True)
        header.setCurrentIndex(index)
        header.edit(index)

    def contextMenuEvent(self, event):
        """Reimplement Qt method."""
        v = QPoint(event.x() - self.table_index.x(), event.y() -
                   self.table_index.y())
        if self.table_index.indexAt(v).isValid():
            self.menu_header_v.popup(event.globalPos())
            event.accept()

    def create_table_level(self):
        """Create the QTableView that will hold the level model."""
        self.table_level = QTableView()
        self.table_level.setEditTriggers(QTableWidget.NoEditTriggers)
        self.table_level.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.table_level.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.table_level.setFrameStyle(QFrame.Plain)
        self.table_level.horizontalHeader().sectionResized.connect(
            self._index_resized)
        self.table_level.verticalHeader().sectionResized.connect(
            self._header_resized)
        self.table_level.setItemDelegate(QItemDelegate())
        self.glayout.addWidget(self.table_level, 0, 0)
        self.table_level.setContentsMargins(0, 0, 0, 0)
        self.table_level.horizontalHeader().sectionClicked.connect(
            self.sortByIndex)

    def create_table_header(self):
        """Create the QTableView that will hold the header model."""
        self.table_header = QTableView()
        self.table_header.verticalHeader().hide()
        self.table_header.setEditTriggers(QTableWidget.NoEditTriggers)
        self.table_header.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.table_header.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.table_header.setHorizontalScrollMode(QTableView.ScrollPerPixel)
        self.table_header.setHorizontalScrollBar(self.hscroll)
        self.table_header.setFrameStyle(QFrame.Plain)
        self.table_header.horizontalHeader().sectionResized.connect(
            self._column_resized)
        self.table_header.setItemDelegate(QItemDelegate())
        self.glayout.addWidget(self.table_header, 0, 1)

    def create_table_index(self):
        """Create the QTableView that will hold the index model."""
        self.table_index = QTableView()
        self.table_index.horizontalHeader().hide()
        self.table_index.setEditTriggers(QTableWidget.NoEditTriggers)
        self.table_index.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.table_index.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.table_index.setVerticalScrollMode(QTableView.ScrollPerPixel)
        self.table_index.setVerticalScrollBar(self.vscroll)
        self.table_index.setFrameStyle(QFrame.Plain)
        self.table_index.verticalHeader().sectionResized.connect(
            self._row_resized)
        self.table_index.setItemDelegate(QItemDelegate())
        self.glayout.addWidget(self.table_index, 1, 0)
        self.table_index.setContentsMargins(0, 0, 0, 0)

    def create_data_table(self):
        """Create the QTableView that will hold the data model."""
        self.dataTable = DataFrameView(
            self,
            self.dataModel,
            self.table_header.horizontalHeader(),
            self.hscroll,
            self.vscroll,
            self.namespacebrowser,
            self.data_function,
            self.readonly
        )
        self.dataTable.verticalHeader().hide()
        self.dataTable.horizontalHeader().hide()
        self.dataTable.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.dataTable.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.dataTable.setHorizontalScrollMode(QTableView.ScrollPerPixel)
        self.dataTable.setVerticalScrollMode(QTableView.ScrollPerPixel)
        self.dataTable.setFrameStyle(QFrame.Plain)
        self.dataTable.setItemDelegate(QItemDelegate())
        self.glayout.addWidget(self.dataTable, 1, 1)
        self.setFocusProxy(self.dataTable)
        self.dataTable.sig_sort_by_column.connect(self._sort_update)
        self.dataTable.sig_fetch_more_columns.connect(self._fetch_more_columns)
        self.dataTable.sig_fetch_more_rows.connect(self._fetch_more_rows)

    def sortByIndex(self, index):
        """Implement a Index sort."""
        self.table_level.horizontalHeader().setSortIndicatorShown(True)
        sort_order = self.table_level.horizontalHeader().sortIndicatorOrder()
        self.table_index.model().sort(index, sort_order)
        self._sort_update()

    def model(self):
        """Get the model of the dataframe."""
        return self._model

    def _column_resized(self, col, old_width, new_width):
        """Update the column width."""
        self.dataTable.setColumnWidth(col, new_width)
        self._update_layout()

    def _row_resized(self, row, old_height, new_height):
        """Update the row height."""
        self.dataTable.setRowHeight(row, new_height)
        self._update_layout()

    def _index_resized(self, col, old_width, new_width):
        """Resize the corresponding column of the index section selected."""
        self.table_index.setColumnWidth(col, new_width)
        self._update_layout()

    def _header_resized(self, row, old_height, new_height):
        """Resize the corresponding row of the header section selected."""
        self.table_header.setRowHeight(row, new_height)
        self._update_layout()

    def _update_layout(self):
        """Set the width and height of the QTableViews and hide rows."""
        h_width = max(self.table_level.verticalHeader().sizeHint().width(),
                      self.table_index.verticalHeader().sizeHint().width())
        self.table_level.verticalHeader().setFixedWidth(h_width)
        self.table_index.verticalHeader().setFixedWidth(h_width)

        last_row = self._model.header_shape[0] - 1
        if last_row < 0:
            hdr_height = self.table_level.horizontalHeader().height()
        else:
            hdr_height = self.table_level.rowViewportPosition(last_row) + \
                         self.table_level.rowHeight(last_row) + \
                         self.table_level.horizontalHeader().height()
            # Check if the header shape has only one row (which display the
            # same info than the horizontal header).
            if last_row == 0:
                self.table_level.setRowHidden(0, True)
                self.table_header.setRowHidden(0, True)
        self.table_header.setFixedHeight(hdr_height)
        self.table_level.setFixedHeight(hdr_height)

        last_col = self._model.header_shape[1] - 1
        if last_col < 0:
            idx_width = self.table_level.verticalHeader().width()
        else:
            idx_width = (
                self.table_level.columnViewportPosition(last_col)
                + self.table_level.columnWidth(last_col)
                + self.table_level.verticalHeader().width()
                # This is necessary to show the separator that allows to resize
                # the index.
                + 5
            )
        self.table_index.setFixedWidth(idx_width)
        self.table_level.setFixedWidth(idx_width)
        self._resizeVisibleColumnsToContents()

    def _reset_model(self, table, model):
        """Set the model in the given table."""
        old_sel_model = table.selectionModel()
        table.setModel(model)
        if old_sel_model:
            del old_sel_model

    def setAutosizeLimitTime(self, limit_ms):
        """Set maximum time to calculate size hint for columns."""
        self._max_autosize_ms = limit_ms

    def setModel(self, model, relayout=True):
        """Set the model for the data, header/index and level views."""
        self._model = model
        sel_model = self.dataTable.selectionModel()
        sel_model.currentColumnChanged.connect(
                self._resizeCurrentColumnToContents)

        # Asociate the models (level, vertical index and horizontal header)
        # with its corresponding view.
        self._reset_model(self.table_level, DataFrameLevelModel(model))
        self._reset_model(self.table_header, DataFrameHeaderModel(model, 0))

        # We use our monospace font for the index so that it matches the one
        # used for data and things look consistent.
        # Fixes issue spyder-ide/spyder#20960
        self._reset_model(
            self.table_index,
            DataFrameHeaderModel(model, 1, use_monospace_font=True)
        )

        # Needs to be called after setting all table models
        if relayout:
            self._update_layout()

    def setCurrentIndex(self, y, x):
        """Set current selection."""
        self.dataTable.selectionModel().setCurrentIndex(
            self.dataTable.model().index(y, x),
            QItemSelectionModel.ClearAndSelect)

    def _sizeHintForColumn(self, table, col, limit_ms=None):
        """Get the size hint for a given column in a table."""
        max_row = table.model().rowCount()
        lm_start = perf_counter()
        lm_row = 64 if limit_ms else max_row
        max_width = self.min_trunc
        for row in range(max_row):
            v = table.sizeHintForIndex(table.model().index(row, col))
            max_width = max(max_width, v.width())
            if row > lm_row:
                lm_now = perf_counter()
                lm_elapsed = (lm_now - lm_start) * 1000
                if lm_elapsed >= limit_ms:
                    break
                lm_row = int((row / lm_elapsed) * limit_ms)
        return max_width

    def _resizeColumnToContents(self, header, data, col, limit_ms):
        """Resize a column by its contents."""
        hdr_width = self._sizeHintForColumn(header, col, limit_ms)
        data_width = self._sizeHintForColumn(data, col, limit_ms)
        if data_width > hdr_width:
            width = min(self.max_width, data_width)
        elif hdr_width > data_width * 2:
            width = max(min(hdr_width, self.min_trunc), min(self.max_width,
                        data_width))
        else:
            width = max(min(self.max_width, hdr_width), self.min_trunc)
        header.setColumnWidth(col, width)

    def _resizeColumnsToContents(self, header, data, limit_ms):
        """Resize all the colummns to its contents."""
        max_col = data.model().columnCount()
        if limit_ms is None:
            max_col_ms = None
        else:
            max_col_ms = limit_ms / max(1, max_col)
        for col in range(max_col):
            self._resizeColumnToContents(header, data, col, max_col_ms)

    def eventFilter(self, obj, event):
        """Override eventFilter to catch resize event."""
        if obj == self.dataTable and event.type() == QEvent.Resize:
            self._resizeVisibleColumnsToContents()
        return False

    def _resizeVisibleColumnsToContents(self):
        """Resize the columns that are in the view."""
        index_column = self.dataTable.rect().topLeft().x()
        start = col = self.dataTable.columnAt(index_column)
        width = self._model.shape[1]
        end = self.dataTable.columnAt(self.dataTable.rect().bottomRight().x())
        end = width if end == -1 else end + 1
        if self._max_autosize_ms is None:
            max_col_ms = None
        else:
            max_col_ms = self._max_autosize_ms / max(1, end - start)
        while col < end:
            resized = False
            if col not in self._autosized_cols:
                self._autosized_cols.add(col)
                resized = True
                self._resizeColumnToContents(self.table_header, self.dataTable,
                                             col, max_col_ms)
            col += 1
            if resized:
                # As we resize columns, the boundary will change
                index_column = self.dataTable.rect().bottomRight().x()
                end = self.dataTable.columnAt(index_column)
                end = width if end == -1 else end + 1
                if max_col_ms is not None:
                    max_col_ms = self._max_autosize_ms / max(1, end - start)

    def _resizeCurrentColumnToContents(self, new_index, old_index):
        """Resize the current column to its contents."""
        if new_index.column() not in self._autosized_cols:
            # Ensure the requested column is fully into view after resizing
            self._resizeVisibleColumnsToContents()
            self.dataTable.scrollTo(new_index)

    def resizeColumnsToContents(self):
        """Resize the columns to its contents."""
        self._autosized_cols = set()
        self._resizeColumnsToContents(self.table_level,
                                      self.table_index, self._max_autosize_ms)
        self._update_layout()

    def show_preferences_dialog(self) -> None:
        """
        Show dialog for setting view options and process user choices.
        """
        # Create dialog using current options
        dialog = PreferencesDialog('dataframe', parent=self)
        dialog.float_format = self.dataModel.get_format_spec()
        dialog.varying_background = self.dataModel.bgcolor_enabled
        dialog.global_algo = not self.dataModel.colum_avg_enabled

        # Show dialog and allow user to interact
        result = dialog.exec_()

        # If user clicked 'OK' then set new options accordingly
        if result == QDialog.Accepted:
            float_format = dialog.float_format

            # This is necessary to handle formatting for integers.
            # Fixes spyder-ide/spyder#22629
            if float_format == 'd':
                float_format = '.0f'

            try:
                format(1.1, float_format)
            except:
                msg = _("Format ({}) is incorrect").format(float_format)
                QMessageBox.critical(self, _("Error"), msg)
            else:
                self.dataModel.set_format_spec(float_format)
                self.set_conf('dataframe_format', float_format)

            self.dataModel.bgcolor(dialog.varying_background)
            if dialog.varying_background:
                self.dataModel.colum_avg(not dialog.global_algo)

    def refresh_editor(self) -> None:
        """
        Refresh data in editor.
        """
        assert self.data_function is not None

        if self.btn_save_and_close and self.btn_save_and_close.isEnabled():
            if not self.ask_for_refresh_confirmation():
                return

        try:
            data = self.data_function()
        except (IndexError, KeyError):
            self.error(_('The variable no longer exists.'))
            return

        if not self.set_data_and_check(data):
            self.error(
                _('The new value cannot be displayed in the dataframe '
                  'editor.')
            )

    def ask_for_refresh_confirmation(self) -> bool:
        """
        Ask user to confirm refreshing the editor.

        This function is to be called if refreshing the editor would overwrite
        changes that the user made previously. The function returns True if
        the user confirms that they want to refresh and False otherwise.
        """
        message = _('Refreshing the editor will overwrite the changes that '
                    'you made. Do you want to proceed?')
        result = QMessageBox.question(
            self,
            _('Refresh dataframe editor?'),
            message
        )
        return result == QMessageBox.Yes

    def error(self, message):
        """An error occurred, closing the dialog box"""
        QMessageBox.critical(self, _("Dataframe editor"), message)
        self.reject()

    def get_value(self):
        """Return modified Dataframe -- this is *not* a copy"""
        # It is import to avoid accessing Qt C++ object as it has probably
        # already been destroyed, due to the Qt.WA_DeleteOnClose attribute
        df = self.dataModel.get_data()
        if self.is_series:
            return df.iloc[:, 0]
        else:
            return df

    def _update_header_size(self):
        """Update the column width of the header."""
        if self.resizeToHeader:
            self.table_header.resizeColumnsToContents()
        column_count = self.table_header.model().columnCount()
        for index in range(0, column_count):
            if index < column_count:
                column_width = self.dataTable.columnWidth(index)
                header_width = self.table_header.columnWidth(index)
                if column_width > header_width:
                    self.table_header.setColumnWidth(index, column_width)
                else:
                    self.dataTable.setColumnWidth(index, header_width)
            else:
                break

    def _update_index_size(self):
        """Update the index's column width."""
        if self.resizeToHeader:
            self.table_level.resizeColumnsToContents()


        column_count = self.table_level.model().columnCount()
        for index in range(0, column_count):
            if index < column_count:
                column_width = self.table_index.columnWidth(index)
                header_width = self.table_level.columnWidth(index)
                if column_width > header_width:
                    self.table_level.setColumnWidth(index, column_width)
                else:
                    self.table_index.setColumnWidth(index, header_width)
            else:
                break

    def _sort_update(self):
        """
        Update the model for all the QTableView objects.

        Uses the model of the dataTable as the base.
        """
        # Update index list calculation
        self.dataModel.recalculate_index()
        self.setModel(self.dataTable.model())

    def _reload(self):
        """
        Reload the model for all the QTableView objects.

        Uses the model of the dataTable as the base.
        """
        # Update index list calculation and reload model
        self.dataModel.recalculate_index()
        self.dataModel.reset()
        self.setModel(self.dataTable.model())

    def _fetch_more_columns(self):
        """Fetch more data for the header (columns)."""
        self.table_header.model().fetch_more()

    def _fetch_more_rows(self):
        """Fetch more data for the index (rows)."""
        self.table_index.model().fetch_more()

    def _update_flag_resize(self):
        self.resizeToHeader = not self.resizeToHeader
        if self.resizeToHeader:
            self.dataTable.resize_columns_action.setText(
                _("Resize columns to headers")
            )
        else:
            self.dataTable.resize_columns_action.setText(
                _("Resize columns to contents")
            )

    @Slot()
    def resize_to_contents(self):
        """"Resize columns to contents"""
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        self.dataTable.resizeColumnsToContents()
        self.dataModel.fetch_more(columns=True)
        self.dataTable.resizeColumnsToContents()
        self._update_header_size()
        self._update_index_size()
        self._update_flag_resize()
        self._update_layout()

        QApplication.restoreOverrideCursor()


# ==============================================================================
# Tests
# ==============================================================================
def test_edit(data, title="", parent=None):
    """Test subroutine"""
    dlg = DataFrameEditor(parent=parent)

    if dlg.setup_and_check(data, title=title):
        dlg.exec_()
        return dlg.get_value()
    else:
        import sys
        sys.exit(1)


def test():
    """DataFrame editor test"""
    from numpy import nan

    if parse(pd.__version__) >= parse('2.0.0'):
        from pandas.testing import assert_frame_equal, assert_series_equal
    else:
        from pandas.util.testing import assert_frame_equal, assert_series_equal

    app = qapplication()                  # analysis:ignore

    df1 = pd.DataFrame(
        [
            [True, "bool"],
            [1+1j, "complex"],
            ['test', "string"],
            [1.11, "float"],
            [1, "int"],
            [np.random.rand(3, 3), "Unkown type"],
            ["Large value", 100],
            ["áéí", "unicode"]
        ],
        index=['a', 'b', nan, nan, nan, 'c', "Test global max", 'd'],
        columns=[nan, 'Type']
    )
    out = test_edit(df1)
    assert_frame_equal(df1, out)

    result = pd.Series([True, "bool"], index=[nan, 'Type'], name='a')
    out = test_edit(df1.iloc[0])
    assert_series_equal(result, out)

    df1 = pd.DataFrame(np.random.rand(100100, 10))
    out = test_edit(df1)
    assert_frame_equal(out, df1)

    series = pd.Series(np.arange(10), name=0)
    out = test_edit(series)
    assert_series_equal(series, out)


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Default tours."""

# Standard library imports
import sys

# Local imports
from spyder.api.translations import _
from spyder.plugins.tours.api import SpyderWidgets as sw
from spyder import __docs_url__


# Constants
QTCONSOLE_LINK = "https://qtconsole.readthedocs.io/en/stable/index.html"
BUTTON_TEXT = ""
if sys.platform != "darwin":
    BUTTON_TEXT = _(
        "Please click on the button below to run some simple code in this "
        "console. This will be useful to show you other important features."
    )

# This test should serve as example of keys to use in the tour frame dicts
TEST_TOUR = [
    {
        'title': "Welcome to Spyder introduction tour",
        'content': "<b>Spyder</b> is an interactive development "
                   "environment. This tip panel supports rich text. <br>"
                   "<br> it also supports image insertion to the right so far",
        'image': 'spyder_about',
    },
    {
        'title': "Widget display",
        'content': ("This shows how a widget is displayed. The tip panel "
                    "is adjusted based on the first widget in the list"),
        'widgets': ['button1'],
        'decoration': ['button2'],
        'interact': True,
    },
    {
        'title': "Widget display",
        'content': ("This shows how a widget is displayed. The tip panel "
                    "is adjusted based on the first widget in the list"),
        'widgets': ['button1'],
        'decoration': ['button1'],
        'interact': True,
    },
    {
        'title': "Widget display",
        'content': ("This shows how a widget is displayed. The tip panel "
                    "is adjusted based on the first widget in the list"),
        'widgets': ['button1'],
        'interact': True,
    },
    {
        'title': "Widget display and highlight",
        'content': "This shows how a highlighted widget looks",
        'widgets': ['button'],
        'decoration': ['button'],
        'interact': False,
    },
]


INTRO_TOUR = [
    {
        'title': _("Welcome to the introduction tour!"),
        'content': _("<b>Spyder</b> is a powerful Interactive "
                     "Development Environment (or IDE) for the Python "
                     "programming language.<br><br>"
                     "Here, we are going to guide you through its most "
                     "important features.<br><br>"
                     "Please use the arrow keys or click on the buttons "
                     "below to move along the tour."),
        'image': 'spyder_about',
    },
    {
        'title': _("Editor"),
        'content': _("This is where you write Python code before "
                     "evaluating it. You can get automatic "
                     "completions while typing, along with calltips "
                     "when calling a function and help when hovering "
                     "over an object."
                     "<br><br>The Editor comes "
                     "with a line number area (highlighted here in red) "
                     "where Spyder shows warnings and syntax errors. "
                     "They can help you to detect potential problems "
                     "before running your code.<br><br>"
                     "You can also set debug breakpoints in the line "
                     "number area by clicking next to "
                     "any non-empty line."),
        'widgets': [sw.editor],
        'decoration': [sw.editor_line_number_area],
    },
    {
        'title': _("IPython Console"),
        'content': _("This is where you can run Python code, either "
                     "from the Editor or interactively. To run the "
                     "current file, press <b>F5</b> by default, "
                     "or press <b>F9</b> to execute the current "
                     "line or selection.<br><br>"
                     "The IPython Console comes with many "
                     "useful features that greatly improve your "
                     "programming workflow, like syntax highlighting, "
                     "autocompletion, plotting and 'magic' commands. "
                     "To learn more, check out the "
                     "<a href=\"{0}\">documentation</a>."
                     "<br><br>{1}").format(QTCONSOLE_LINK, BUTTON_TEXT),
        'widgets': [sw.ipython_console],
        'run': [
            "test_list_tour = [1, 2, 3, 4, 5]",
            "test_dict_tour = {'a': 1, 'b': 2}",
        ],
    },
    {
        'title': _("Variable Explorer"),
        'content': _("In this pane you can view and edit the variables "
                     "generated during the execution of a program, or "
                     "those entered directly in the "
                     "IPython Console.<br><br>"
                     "If you ran the code in the previous step, "
                     "the Variable Explorer will show "
                     "the list and dictionary objects it generated. "
                     "By double-clicking any variable, "
                     "a new window will be opened where you "
                     "can inspect and modify their contents."),
        'widgets': [sw.variable_explorer],
        'interact': True,
    },
    {
        'title': _("Help"),
        'content': _("This pane displays documentation of the "
                     "functions, classes, methods or modules you are "
                     "currently using in the Editor or the "
                     "IPython Console."
                     "<br><br>To use it, press <b>Ctrl+I</b> "
                     "(<b>Cmd-I</b> on macOS) with the text cursor "
                     "in or next to the object you want help on."),
        'widgets': [sw.help_plugin],
        'interact': True,
    },
    {
        'title': _("Plots"),
        'content': _("This pane shows the figures and images created "
                     "during your code execution. It allows you to browse, "
                     "zoom, copy, and save the generated plots."),
        'widgets': [sw.plots_plugin],
        'interact': True,
    },
    {
        'title': _("Files"),
        'content': _("This pane lets you browse the files and "
                     "directories on your computer.<br><br>"
                     "You can open any file in its "
                     "corresponding application by double-clicking it, "
                     "and supported file types will be opened right "
                     "inside of Spyder.<br><br>"
                     "The Files pane also allows you to copy one or "
                     "many absolute or relative paths, automatically "
                     "formatted as Python strings or lists, and perform "
                     "a variety of other file operations."),
        'widgets': [sw.file_explorer],
        'interact': True,
    },
    {
        'title': _("History Log"),
        'content': _("This pane records all the commands and code run "
                     "in any IPython console, allowing you to easily "
                     "retrace your steps for reproducible research."),
        'widgets': [sw.history_log],
        'interact': True,
    },
    {
        'title': _("Find"),
        'content': _("The Find pane allows you to search for text in a "
                     "given directory and navigate through all the found "
                     "occurrences."),
        'widgets': [sw.find_plugin],
        'interact': True,
    },
    {
        'title': _("Profiler"),
        'content': _("The Profiler helps you optimize your code by "
                     "determining the run time and number of calls for "
                     "every function and method used in a file. It also "
                     "allows you to save and compare your results between "
                     "runs."),
        'widgets': [sw.profiler],
        'interact': True,
    },
    {
        'title': _("Code Analysis"),
        'content': _("The Code Analysis helps you improve the quality of "
                     "your programs by detecting style issues, bad practices "
                     "and potential bugs."),
        'widgets': [sw.code_analysis],
        'interact': True
    },
    {
        'title': _("The end"),
        'content': _('You have reached the end of our tour and are '
                     'ready to start using Spyder! For more '
                     'information, check out our '
                     '<a href="{}">documentation</a>.'
                     '<br><br>').format(__docs_url__),
        'image': 'spyder_about'
    },
]


FEAT30 = [
    {
        'title': _("New features in Spyder 3.0"),
        'content': _("<b>Spyder</b> is an interactive development "
                     "environment based on bla"),
        'image': 'spyder_about',
    },
    {
        'title': _("Welcome to Spyder introduction tour"),
        'content': _("Spyder is an interactive development environment "
                     "based on bla"),
        'widgets': ['variableexplorer'],
    },
]


class TourIdentifiers:
    IntroductionTour = "introduction_tour"
    TestTour = "test_tour"



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder path manager."""

# Standard library imports
from collections import OrderedDict
import os
import os.path as osp
import sys

# Third party imports
from qtpy import PYSIDE2
from qtpy.compat import getexistingdirectory
from qtpy.QtCore import QSize, Qt, Signal, Slot
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QStackedWidget,
    QVBoxLayout,
)
from requests.structures import CaseInsensitiveDict

# Local imports
from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.plugins.pythonpath.utils import check_path, get_system_pythonpath
from spyder.utils.environ import (
    get_user_environment_variables,
    get_user_env,
    set_user_env,
)
from spyder.utils.misc import getcwd_or_home
from spyder.utils.stylesheet import (
    AppStyle,
    MAC,
    PANES_TOOLBAR_STYLESHEET,
    WIN
)
from spyder.widgets.emptymessage import EmptyMessageWidget


class PathManagerToolbuttons:
    MoveTop = 'move_top'
    MoveUp = 'move_up'
    MoveDown = 'move_down'
    MoveToBottom = 'move_to_bottom'
    AddPath = 'add_path'
    RemovePath = 'remove_path'
    ImportPaths = 'import_paths'
    ExportPaths = 'export_paths'
    Prioritize = 'prioritize'


class PathManager(QDialog, SpyderWidgetMixin):
    """Path manager dialog."""

    redirect_stdio = Signal(bool)
    sig_path_changed = Signal(object, object, bool)

    # This is required for our tests
    CONF_SECTION = 'pythonpath_manager'

    def __init__(self, parent, sync=True):
        """Path manager dialog."""
        if not PYSIDE2:
            super().__init__(parent, class_parent=parent)
        else:
            QDialog.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=parent)

        # Style
        # NOTE: This needs to be here so all buttons are styled correctly
        self.setStyleSheet(self._stylesheet)

        self.last_path = getcwd_or_home()

        # Widgets
        self.add_button = None
        self.remove_button = None
        self.movetop_button = None
        self.moveup_button = None
        self.movedown_button = None
        self.movebottom_button = None
        self.export_button = None
        self.prioritize_button = None
        self.user_header = None
        self.project_header = None
        self.system_header = None
        self.headers = []
        self.selection_widgets = []
        self.right_buttons = self._setup_right_toolbar()
        self.listwidget = QListWidget(self)
        self.bbox = SpyderDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.button_ok = self.bbox.button(QDialogButtonBox.Ok)

        # Create a loading message
        self.loading_pane = EmptyMessageWidget(
            parent=self,
            text=_("Retrieving environment variables..."),
            bottom_stretch=1,
            spinner=True,
        )

        # Create a QStackedWidget
        self.stacked_widget = QStackedWidget()
        self.stacked_widget.addWidget(self.listwidget)
        self.stacked_widget.addWidget(self.loading_pane)
        self.stacked_widget.setCurrentWidget(self.listwidget)

        # Widget setup
        self.setWindowTitle(_("PYTHONPATH manager"))
        self.setWindowIcon(self.create_icon('pythonpath'))
        self.resize(500, 400)
        self.export_button.setVisible(os.name == 'nt' and sync)

        # Description
        description = QLabel(
            _("The paths listed below will be passed to the IPython console "
              "and to the Editor as additional locations to search for Python "
              "modules.")
        )
        description.setWordWrap(True)

        # Buttons layout
        buttons_layout = QVBoxLayout()
        self._add_buttons_to_layout(self.right_buttons, buttons_layout)
        buttons_layout.addStretch(1)

        # Middle layout
        middle_layout = QHBoxLayout()
        middle_layout.setContentsMargins(4 if WIN else 5, 0, 0, 0)
        middle_layout.addWidget(self.stacked_widget)
        middle_layout.addLayout(buttons_layout)

        # Widget layout
        layout = QVBoxLayout()
        layout.addWidget(description)
        layout.addSpacing(2 * AppStyle.MarginSize)
        layout.addLayout(middle_layout)
        layout.addSpacing((-1 if MAC else 2) * AppStyle.MarginSize)
        layout.addWidget(self.bbox)
        self.setLayout(layout)

        # Signals
        self.listwidget.currentRowChanged.connect(lambda x: self.refresh())
        self.listwidget.itemChanged.connect(lambda x: self.refresh())
        self.bbox.accepted.connect(self.accept)
        self.bbox.rejected.connect(self.reject)

        # Attributes
        self.project_path = None
        self.user_paths = None
        self.system_paths = None
        self.prioritize = None

    # ---- Private methods
    # -------------------------------------------------------------------------
    def _add_buttons_to_layout(self, widgets, layout):
        """Helper to add buttons to its layout."""
        for widget in widgets:
            layout.addWidget(widget)

    def _setup_right_toolbar(self):
        """Create top toolbar and actions."""
        self.movetop_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveTop,
            text=_("Move path to the top"),
            icon=self.create_icon('2uparrow'),
            triggered=lambda: self.move_to(absolute=0))
        self.moveup_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveUp,
            tip=_("Move path up"),
            icon=self.create_icon('1uparrow'),
            triggered=lambda: self.move_to(relative=-1))
        self.movedown_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveDown,
            tip=_("Move path down"),
            icon=self.create_icon('1downarrow'),
            triggered=lambda: self.move_to(relative=1))
        self.movebottom_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveToBottom,
            text=_("Move path to the bottom"),
            icon=self.create_icon('2downarrow'),
            triggered=lambda: self.move_to(absolute=1))
        self.add_button = self.create_toolbutton(
            PathManagerToolbuttons.AddPath,
            tip=_('Add path'),
            icon=self.create_icon('edit_add'),
            triggered=lambda x: self.add_path())
        self.remove_button = self.create_toolbutton(
            PathManagerToolbuttons.RemovePath,
            tip=_('Remove path'),
            icon=self.create_icon('editclear'),
            triggered=lambda x: self.remove_path())
        self.import_button = self.create_toolbutton(
            PathManagerToolbuttons.ImportPaths,
            tip=_('Import from PYTHONPATH environment variable'),
            icon=self.create_icon('fileimport'),
            triggered=self.import_pythonpath)
        self.export_button = self.create_toolbutton(
            PathManagerToolbuttons.ExportPaths,
            icon=self.create_icon('fileexport'),
            triggered=self.export_pythonpath,
            tip=_("Export to PYTHONPATH environment variable"))
        self.prioritize_button = self.create_toolbutton(
            PathManagerToolbuttons.Prioritize,
            option='prioritize',
            triggered=self.refresh,
        )
        self.prioritize_button.setCheckable(True)

        self.selection_widgets = [self.movetop_button, self.moveup_button,
                                  self.movedown_button, self.movebottom_button]
        return (
            [self.add_button, self.remove_button] +
            self.selection_widgets + [self.import_button, self.export_button] +
            [self.prioritize_button]
        )

    def _create_item(self, path, active):
        """Helper to create a new list item."""
        item = QListWidgetItem(path)

        item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
        item.setCheckState(Qt.Checked if active else Qt.Unchecked)

        return item

    def _create_header(self, text):
        """Create a header for a given path section."""
        header_item = QListWidgetItem()
        header_widget = QLabel(text)

        # Disable item so we can remove its background color
        header_item.setFlags(header_item.flags() & ~Qt.ItemIsEnabled)

        # Header is centered
        header_widget.setAlignment(Qt.AlignHCenter)

        # Make header appear in bold
        font = header_widget.font()
        font.setBold(True)
        header_widget.setFont(font)

        # Increase height to make header stand over paths
        fm = QFontMetrics(font)
        header_item.setSizeHint(
            QSize(20, fm.capHeight() + 6 * AppStyle.MarginSize)
        )

        return header_item, header_widget

    @property
    def _stylesheet(self):
        """Style for the list of paths"""
        # This is necessary to match the buttons style with the rest of Spyder
        toolbar_stylesheet = PANES_TOOLBAR_STYLESHEET.get_copy()
        css = toolbar_stylesheet.get_stylesheet()

        css.QListView.setValues(
            padding=f"{AppStyle.MarginSize + 1}px"
        )

        css["QListView::item"].setValues(
            padding=f"{AppStyle.MarginSize + (1 if WIN else 0)}px"
        )

        css["QListView::item:disabled"].setValues(
            backgroundColor="transparent"
        )

        return css.toString()

    def _setup_system_paths(self, paths):
        """Add system paths, creating system header if necessary"""
        if not paths:
            return

        if not self.system_header:
            self.system_header, system_widget = (
                self._create_header(_("System PYTHONPATH"))
            )
            self.headers.append(self.system_header)
            self.listwidget.addItem(self.system_header)
            self.listwidget.setItemWidget(self.system_header, system_widget)

        for path, active in paths.items():
            item = self._create_item(path, active)
            self.listwidget.addItem(item)

    # ---- Public methods
    # -------------------------------------------------------------------------
    @property
    def editable_bottom_row(self):
        """Maximum bottom row count that is editable."""
        bottom_row = 0

        if self.project_header:
            bottom_row += len(self.project_path) + 1
        if self.user_header:
            bottom_row += len(self.get_user_paths())

        return bottom_row

    @property
    def editable_top_row(self):
        """Maximum top row count that is editable."""
        top_row = 0

        if self.project_header:
            top_row += len(self.project_path) + 1
        if self.user_header:
            top_row += 1

        return top_row

    def setup(self):
        """Populate list widget."""
        self.listwidget.clear()
        self.headers.clear()
        self.project_header = None
        self.user_header = None
        self.system_header = None

        # Project path
        if self.project_path:
            self.project_header, project_widget = (
                self._create_header(_("Project path"))
            )
            self.headers.append(self.project_header)
            self.listwidget.addItem(self.project_header)
            self.listwidget.setItemWidget(self.project_header, project_widget)

            for path, active in self.project_path.items():
                item = self._create_item(path, active)

                # Project path should not be editable
                item.setFlags(Qt.NoItemFlags | Qt.ItemIsUserCheckable)
                item.setCheckState(Qt.Checked)

                self.listwidget.addItem(item)

        # Paths added by the user
        if self.user_paths:
            self.user_header, user_widget = (
                self._create_header(_("User paths"))
            )
            self.headers.append(self.user_header)
            self.listwidget.addItem(self.user_header)
            self.listwidget.setItemWidget(self.user_header, user_widget)

            for path, active in self.user_paths.items():
                item = self._create_item(path, active)
                self.listwidget.addItem(item)

        # System paths
        self._setup_system_paths(self.system_paths)

        # Prioritize
        self.prioritize_button.setChecked(self.prioritize)

        self.listwidget.setCurrentRow(0)
        self.refresh()

    @Slot()
    def export_pythonpath(self):
        """
        Export to Spyder's PYTHONPATH Manager paths to HKCU PYTHONPATH
        environment variable (Windows only).

        If the user chooses to keep the inactive system paths, then the active
        user paths and all system paths are exported, in that order.

        If the user chooses not to keep the inactive system paths, then the
        active user paths and active system paths are exported, in that order.

        With either choice, inactive user paths are retained as inactive user
        paths in the PYTHONPATH Manager and the system paths reflect the
        exported paths, retaining their active state.
        """
        answer = QMessageBox.question(
            self,
            _("Export"),
            _("This will export Spyder's path list to the "
              "<b>PYTHONPATH</b> environment variable for the current user, "
              "allowing you to run your Python modules outside Spyder "
              "without having to configure sys.path. "
              "<br><br>"
              "Do you want to keep the inactive system paths?"),
            QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel
        )

        if answer == QMessageBox.Cancel:
            return

        user_paths = self.get_user_paths()
        active_user_paths = OrderedDict(
            {p: v for p, v in user_paths.items() if v}
        )
        new_user_paths = OrderedDict(
            {p: v for p, v in user_paths.items() if not v}
        )

        system_paths = self.get_system_paths()

        if answer == QMessageBox.No:
            # Only use active system paths
            system_paths = OrderedDict(
                {p: v for p, v in system_paths.items() if v}
            )

        new_system_paths = active_user_paths.copy()
        for k, v in system_paths.items():
            new_system_paths.setdefault(k, v)

        env = get_user_env()
        if os.name == "nt":
            env = CaseInsensitiveDict(env)
        env["PYTHONPATH"] = list(new_system_paths.keys())

        set_user_env(dict(env), parent=self)

        self.update_paths(
            user_paths=new_user_paths, system_paths=new_system_paths
        )

    def get_user_paths(self):
        """Get current user paths as displayed on listwidget."""
        paths = OrderedDict()

        if self.user_header is None:
            return paths

        start = self.listwidget.row(self.user_header) + 1
        stop = self.listwidget.count()
        if self.system_header is not None:
            stop = self.listwidget.row(self.system_header)

        for row in range(start, stop):
            item = self.listwidget.item(row)
            paths.update({item.text(): item.checkState() == Qt.Checked})

        return paths

    def get_system_paths(self):
        """Get current system paths as displayed on listwidget."""
        paths = OrderedDict()

        if self.system_header is None:
            return paths

        start = self.listwidget.row(self.system_header) + 1
        for row in range(start, self.listwidget.count()):
            item = self.listwidget.item(row)
            paths.update({item.text(): item.checkState() == Qt.Checked})

        return paths

    def update_paths(
        self,
        project_path=None,
        user_paths=None,
        system_paths=None,
        prioritize=None
    ):
        """
        Update path attributes.

        These attributes should only be set in this method and upon activating
        the dialog. They should remain fixed while the dialog is active and are
        used to compare with what is shown in the listwidget in order to detect
        changes.
        """
        if project_path is not None:
            self.project_path = project_path
        if user_paths is not None:
            self.user_paths = user_paths
        if system_paths is not None:
            self.system_paths = system_paths
        if prioritize is not None:
            self.prioritize = prioritize

        self.setup()

    def refresh(self):
        """Refresh toolbar widgets."""
        current_item = self.listwidget.currentItem()
        enabled = current_item is not None
        for widget in self.selection_widgets:
            widget.setEnabled(enabled)

        # Main variables
        row = self.listwidget.currentRow()
        disable_widgets = []

        # Move up/top disabled for less than top editable item.
        if row <= self.editable_top_row:
            disable_widgets.extend([self.movetop_button, self.moveup_button])

        # Move down/bottom disabled for bottom item
        if row == self.editable_bottom_row:
            disable_widgets.extend([self.movebottom_button,
                                    self.movedown_button])

        # Disable almost all buttons on headers or system PYTHONPATH
        if current_item in self.headers or row > self.editable_bottom_row:
            disable_widgets.extend(
                [self.movetop_button, self.moveup_button,
                 self.movebottom_button, self.movedown_button]
            )

        for widget in disable_widgets:
            widget.setEnabled(False)

        # Enable remove button only for user paths
        self.remove_button.setEnabled(
            current_item not in self.headers
            and (self.editable_top_row <= row <= self.editable_bottom_row)
        )

        if self.prioritize_button.isChecked():
            self.prioritize_button.setIcon(self.create_icon('prepend'))
            self.prioritize_button.setToolTip(
                _("Paths are prepended to sys.path")
            )
        else:
            self.prioritize_button.setIcon(self.create_icon('append'))
            self.prioritize_button.setToolTip(
                _("Paths are appended to sys.path")
            )

        self.export_button.setEnabled(self.listwidget.count() > 0)

        # Ok button only enabled if actual changes occur
        self.button_ok.setEnabled(
            self.user_paths != self.get_user_paths()
            or self.system_paths != self.get_system_paths()
            or self.prioritize != self.prioritize_button.isChecked()
        )

    @Slot()
    def add_path(self, directory=None):
        """
        Add path to list widget.

        If `directory` is provided, the folder dialog is overridden.
        """
        if directory is None:
            self.redirect_stdio.emit(False)
            directory = getexistingdirectory(self, _("Select directory"),
                                             self.last_path)
            self.redirect_stdio.emit(True)
            if not directory:
                return

        directory = osp.abspath(directory)
        self.last_path = directory

        if directory in self.get_user_paths():
            # Always take the last item to avoid retrieving the project path
            item = self.listwidget.findItems(directory, Qt.MatchExactly)[-1]
            item.setCheckState(Qt.Checked)
            answer = QMessageBox.question(
                self,
                _("Add path"),
                _("This directory is already included in the list."
                  "<br> "
                  "Do you want to move it to the top of the list?"),
                QMessageBox.Yes | QMessageBox.No)

            if answer == QMessageBox.Yes:
                item = self.listwidget.takeItem(self.listwidget.row(item))
                self.listwidget.insertItem(self.editable_top_row, item)
                self.listwidget.setCurrentRow(self.editable_top_row)
        else:
            if check_path(directory):
                if not self.user_header:
                    self.user_header, user_widget = (
                        self._create_header(_("User paths"))
                    )
                    self.headers.append(self.user_header)

                    if self.editable_top_row > 0:
                        header_row = self.editable_top_row - 1
                    else:
                        header_row = 0
                    self.listwidget.insertItem(header_row, self.user_header)
                    self.listwidget.setItemWidget(
                        self.user_header, user_widget
                    )

                # Add new path
                item = self._create_item(directory, True)
                self.listwidget.insertItem(self.editable_top_row, item)
                self.listwidget.setCurrentRow(self.editable_top_row)
            else:
                answer = QMessageBox.warning(
                    self,
                    _("Add path"),
                    _("This directory cannot be added to the path!"
                      "<br><br>"
                      "If you want to set a different Python interpreter, "
                      "please go to <tt>Preferences > Main interpreter</tt>"
                      "."),
                    QMessageBox.Ok)

        # Widget moves to back and loses focus on macOS,
        # see spyder-ide/spyder#20808
        if sys.platform == 'darwin':
            self.activateWindow()
            self.raise_()
            self.setFocus()

        self.refresh()

    @Slot()
    def remove_path(self, force=False):
        """
        Remove path from list widget.

        If `force` is True, the message box is overridden.
        """
        if self.listwidget.currentItem():
            if not force:
                answer = QMessageBox.warning(
                    self,
                    _("Remove path"),
                    _("Do you really want to remove the selected path?"),
                    QMessageBox.Yes | QMessageBox.No)

            if force or answer == QMessageBox.Yes:
                # Remove selected item from view
                self.listwidget.takeItem(self.listwidget.currentRow())

                # Remove user header if there are no more user paths
                if len(self.get_user_paths()) == 0:
                    self.listwidget.takeItem(
                        self.listwidget.row(self.user_header)
                    )
                    self.headers.remove(self.user_header)
                    self.user_header = None

                # Refresh widget
                self.refresh()

    @Slot()
    def import_pythonpath(self):
        future = get_user_environment_variables()
        future.connect(self._import_pythonpath)
        self.stacked_widget.setCurrentWidget(self.loading_pane)

    @AsyncDispatcher.QtSlot
    def _import_pythonpath(self, future):
        """Import PYTHONPATH from environment."""
        current_system_paths = self.get_system_paths()
        system_paths = get_system_pythonpath(future.result())

        # Inherit active state from current system paths
        system_paths = OrderedDict(
            {p: current_system_paths.get(p, True) for p in system_paths}
        )

        # Remove system paths
        if self.system_header:
            header_row = self.listwidget.row(self.system_header)
            for row in range(self.listwidget.count(), header_row, -1):
                self.listwidget.takeItem(row)

            # Also remove system header
            if not system_paths:
                self.listwidget.takeItem(header_row)
                self.headers.remove(self.system_header)
                self.system_header = None

        self._setup_system_paths(system_paths)

        self.stacked_widget.setCurrentWidget(self.listwidget)

        self.refresh()

    def move_to(self, absolute=None, relative=None):
        """Move items of list widget."""
        index = self.listwidget.currentRow()
        if absolute is not None:
            if absolute:
                new_index = self.editable_bottom_row
            else:
                new_index = self.editable_top_row
        else:
            new_index = index + relative

        new_index = max(1, min(self.editable_bottom_row, new_index))
        item = self.listwidget.takeItem(index)
        self.listwidget.insertItem(new_index, item)
        self.listwidget.setCurrentRow(new_index)

        self.refresh()

    def current_row(self):
        """Returns the current row of the list."""
        return self.listwidget.currentRow()

    def set_current_row(self, row):
        """Set the current row of the list."""
        self.listwidget.setCurrentRow(row)

    def row_check_state(self, row):
        """Return the checked state for item in row."""
        item = self.listwidget.item(row)
        return item.checkState()

    def set_row_check_state(self, row, value):
        """Set the current checked state for item in row."""
        item = self.listwidget.item(row)
        item.setCheckState(value)

    def count(self):
        """Return the number of items."""
        return self.listwidget.count()

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def accept(self):
        """Override Qt method."""
        self.sig_path_changed.emit(
            self.get_user_paths(),
            self.get_system_paths(),
            self.prioritize_button.isChecked()
        )
        super().accept()


def test():
    """Run path manager test."""
    from spyder.utils.qthelpers import qapplication

    _ = qapplication()
    dlg = PathManager(
        None,
    )
    dlg.update_paths(
        user_paths={p: True for p in sys.path[1:-2]},
        project_path={p: True for p in sys.path[:1]},
        system_paths={p: True for p in sys.path[-2:]},
        prioritize=False
    )

    def callback(user_paths, system_paths, prioritize):
        sys.stdout.write(f"Prioritize: {prioritize}")
        sys.stdout.write("\n---- User paths ----\n")
        sys.stdout.write(
            '\n'.join([f'{k}: {v}' for k, v in user_paths.items()])
        )
        sys.stdout.write("\n---- System paths ----\n")
        sys.stdout.write(
            '\n'.join([f'{k}: {v}' for k, v in system_paths.items()])
        )
        sys.stdout.write('\n')

    dlg.sig_path_changed.connect(callback)
    sys.exit(dlg.exec_())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Shortcut configuration page."""

# Third party imports
from qtpy.QtWidgets import QHBoxLayout, QLabel, QMessageBox, QVBoxLayout

# Local imports
from spyder.api.preferences import PluginConfigPage
from spyder.api.translations import _
from spyder.plugins.shortcuts.widgets.table import (
    ShortcutFinder,
    ShortcutsTable,
)
from spyder.utils.icon_manager import ima
from spyder.utils.stylesheet import AppStyle


class ShortcutsConfigPage(PluginConfigPage):
    APPLY_CONF_PAGE_SETTINGS = True

    def setup_page(self):
        # Widgets
        self.table = ShortcutsTable(self)
        finder = ShortcutFinder(self.table, self.table.set_regex)
        reset_btn = self.create_button(
            icon=ima.icon("restart"),
            callback=self.reset_to_default,
            tooltip=_("Reset all shortcuts to default values"),
        )
        top_label = QLabel(
            _("Customize a shortcut by double-clicking on its entry below.")
        )

        # Widget setup
        self.table.finder = finder
        self.table.set_shortcut_data(self.plugin.get_shortcut_data())
        self.table.load_shortcuts()
        self.table.finder.setPlaceholderText(_("Type to search"))
        top_label.setWordWrap(True)

        # Layout
        hlayout = QHBoxLayout()
        hlayout.addWidget(finder)
        hlayout.addWidget(reset_btn)

        vlayout = QVBoxLayout()
        vlayout.addWidget(top_label)
        vlayout.addWidget(self.table)
        vlayout.addSpacing(AppStyle.MarginSize)
        vlayout.addLayout(hlayout)
        self.setLayout(vlayout)

        self.setTabOrder(self.table, finder)
        self.setTabOrder(finder, reset_btn)

        # Signals
        self.table.proxy_model.dataChanged.connect(
            lambda i1, i2, roles, opt='', sect='': self.has_been_modified(
                sect, opt))

    def check_settings(self):
        self.table.check_shortcuts()

    def reset_to_default(self, force=False):
        """Reset all shortcuts to default values after confirmation."""
        if not force:
            reset = QMessageBox.warning(
                self,
                _("Reset all shortcuts"),
                _(
                    "Do you want to reset all shortcuts to their default "
                    "values?"
                ),
                QMessageBox.Yes | QMessageBox.No,
            )

            if reset == QMessageBox.No:
                return

        self.plugin.reset_shortcuts()
        self.plugin.apply_shortcuts()
        self.table.load_shortcuts()
        self.load_from_conf()
        self.set_modified(False)

    def apply_settings(self, options):
        self.table.save_shortcuts()
        self.plugin.apply_shortcuts()
        self.plugin.apply_conf(options)



# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Widgets to extend Spyder through its API.
"""


class PluginMainWidgetWidgets:
    """Basic widgets any :class:`~spyder.api.plugins.SpyderDockablePlugin` has."""

    CornerWidget: str = "corner_widget"
    """Right-corner pane toolbar buttons, just left of the options button."""

    MainToolbar: str = "main_toolbar_widget"
    """The primary pane toolbar, left-aligned."""

    OptionsToolButton: str = "options_button_widget"
    """The pane's "hamburger menu", on the very right of the toolbar area."""

    Spinner: str = "spinner_widget"
    """An optional progress spinner widget in the toolbar."""


class PluginMainWidgetActions:
    """Common menu actions for :class:`~spyder.api.plugins.SpyderDockablePlugin`\\s."""

    ClosePane: str = "close_pane"
    """Close the plugin's pane."""

    DockPane: str = "dock_pane"
    """Re-dock a popped-out pane to the Spyder main window."""

    UndockPane: str = "undock_pane"
    """Pop out a plugin's pane into a separate floating window."""

    LockUnlockPosition: str = "lock_unlock_position"
    """Toggle whether a pane can be freely moved around the Spyder window."""



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Language Server Protocol configuration tabs.
"""

# Third party imports
from qtpy.QtWidgets import QGroupBox, QLabel, QVBoxLayout

# Local imports
from spyder.api.preferences import SpyderPreferencesTab
from spyder.api.translations import _


class FormattingConfigTab(SpyderPreferencesTab):
    """Code formatting tab."""

    TITLE = _('Code formatting')

    def __init__(self, parent):
        super().__init__(parent)
        newcb = self.create_checkbox

        pep_url = (
            '<a href="https://www.python.org/dev/peps/pep-0008">PEP 8</a>'
        )

        self.code_style_max_line_length = self.create_spinbox(
            _("Maximum allowed line length:"), None,
            'flake8/max_line_length', min_=10, max_=500, step=1,
            tip=_("Default is 79"))

        vertical_line_box = newcb(
            _("Show vertical line at that length"), 'edge_line',
            section='editor')

        # Maximum allowed line length layout
        line_length_group = QGroupBox(_("Line length"))
        line_length_layout = QVBoxLayout()
        line_length_layout.addWidget(self.code_style_max_line_length)
        line_length_layout.addWidget(vertical_line_box)
        line_length_group.setLayout(line_length_layout)

        # Code formatting label
        autopep8_url = (
            "<a href='https://github.com/hhatto/autopep8'>Autopep8</a>"
        )
        yapf_url = (
            "<a href='https://github.com/google/yapf'>Yapf</a>"
        )
        black_url = (
            "<a href='https://black.readthedocs.io/en/stable'>Black</a>"
        )
        code_fmt_label = QLabel(
            _("Spyder can use {0} or {1} to format your code for "
              "conformance to the {2} convention.").format(
                  autopep8_url, black_url, pep_url))
        code_fmt_label.setOpenExternalLinks(True)
        code_fmt_label.setWordWrap(True)

        # Code formatting providers
        code_fmt_provider = self.create_combobox(
            _("Choose the code formatting provider: "),
            (("autopep8", 'autopep8'),
             ("black", 'black')),
            'formatting')

        # Autoformat on save
        format_on_save_box = newcb(
            _("Autoformat files on save"),
            'format_on_save',
            tip=_("If enabled, autoformatting will take place when "
                  "saving a file"))

        # Code formatting layout
        code_fmt_group = QGroupBox(_("Code formatting"))
        code_fmt_layout = QVBoxLayout()
        code_fmt_layout.addWidget(code_fmt_label)
        code_fmt_layout.addWidget(code_fmt_provider)
        code_fmt_layout.addWidget(format_on_save_box)
        code_fmt_group.setLayout(code_fmt_layout)

        code_style_fmt_layout = QVBoxLayout()
        code_style_fmt_layout.addWidget(code_fmt_group)
        code_style_fmt_layout.addWidget(line_length_group)
        self.setLayout(code_style_fmt_layout)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Switcher Main Widget."""


# Third party imports
from qtpy.QtCore import (
    QEvent,
    QItemSelectionModel,
    QModelIndex,
    QObject,
    Qt,
    Signal,
    Slot,
)
from qtpy.QtGui import QStandardItemModel
from qtpy.QtWidgets import (QAbstractItemView, QDialog, QLineEdit,
                            QListView, QListWidgetItem, QStyle,
                            QVBoxLayout)
from superqt.utils import qdebounced, signals_blocked

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.plugins.switcher.widgets.proxymodel import SwitcherProxyModel
from spyder.plugins.switcher.widgets.item import (
    SwitcherItem, SwitcherSeparatorItem)
from spyder.utils.palette import SpyderPalette
from spyder.widgets.helperwidgets import HTMLDelegate
from spyder.utils.stringmatching import get_search_scores
from spyder.plugins.switcher.utils import clean_string


class KeyPressFilter(QObject):
    """Use with `installEventFilter` to get up/down arrow key press signal."""

    sig_up_key_pressed = Signal()
    sig_down_key_pressed = Signal()
    sig_enter_key_pressed = Signal()

    def eventFilter(self, src, e):
        """Override Qt eventFilter."""
        if e.type() == QEvent.KeyPress:
            if e.key() == Qt.Key_Up:
                self.sig_up_key_pressed.emit()
                return True
            elif e.key() == Qt.Key_Down:
                self.sig_down_key_pressed.emit()
                return True
            elif (e.key() == Qt.Key_Return):
                self.sig_enter_key_pressed.emit()
                return True
        return super().eventFilter(src, e)


class SwitcherDelegate(HTMLDelegate):
    """
    This delegate allows the list view of the switcher to look like it has
    the focus, even when its focus policy is set to Qt.NoFocus.
    """

    def paint(self, painter, option, index):
        """
        Override Qt method to force this delegate to look active at all times.
        """
        option.state |= QStyle.State_Active
        super().paint(painter, option, index)


class Switcher(QDialog, SpyderFontsMixin):
    """
    A multi purpose switcher.

    Example
    -------
      SwitcherItem:      [title description    <shortcut> section]
      SwitcherItem:      [title description    <shortcut> section]
      SwitcherSeparator: [---------------------------------------]
      SwitcherItem:      [title description    <shortcut> section]
      SwitcherItem:      [title description    <shortcut> section]
    """

    # --- Signals
    sig_rejected = Signal()
    """
    This signal is emitted when the plugin is dismissed.
    """

    sig_item_changed = Signal(object)
    """
    This signal is emitted when the plugin current item changes.
    """

    sig_item_selected = Signal(object, str, str)
    """
    This signal is emitted when an item is selected from the switcher plugin
    list.

    Parameters
    ----------
    item: object
        The current selected item from the switcher list (QStandardItem).
    mode: str
        The current selected mode (open files "", symbol "@" or line ":").
    search_text: str
        Cleaned search/filter text.
    """

    sig_mode_selected = Signal(str)
    """
    This signal is emitted when a mode is selected.

    Parameters
    ----------
    mode: str
        The selected mode (open files "", symbol "@" or line ":").
    """

    sig_search_text_available = Signal(str)
    """
    This signal is emitted when the user stops typing in the filter line edit.

    Parameters
    ----------
    search_text: str
        The current search text.
    """

    _MAX_NUM_ITEMS = 15
    _MIN_WIDTH = 580
    _MIN_HEIGHT = 200
    _MAX_HEIGHT = 390
    _ITEM_WIDTH = _MIN_WIDTH - 20

    def __init__(self, parent, help_text=None):
        """Multi purpose switcher."""
        super().__init__(parent)

        # Attributes
        self._modes = {}
        self._mode_on = ''

        font_size = self.get_font(SpyderFontType.Interface).pointSize()
        self._item_styles = {
            'title_color': SpyderPalette.COLOR_TEXT_1,
            'description_color': SpyderPalette.COLOR_TEXT_1,
            'section_color': SpyderPalette.TIP_TITLE_COLOR,
            'shortcut_color': SpyderPalette.COLOR_TEXT_1,
            'title_font_size': font_size,
            'description_font_size': font_size,
            'section_font_size': font_size,
            'shortcut_font_size': font_size,
        }
        self._item_separator_styles = {
            'color': SpyderPalette.COLOR_TEXT_1,
            'font_size': font_size,
        }

        # Widgets
        self.edit = QLineEdit(self)
        self.list = QListView(self)
        self.model = QStandardItemModel(self.list)
        self.proxy = SwitcherProxyModel(self.list)
        self.filter = KeyPressFilter()

        # Widgets setup
        self.setWindowFlags(Qt.Popup | Qt.FramelessWindowHint)
        self.setMaximumHeight(self._MAX_HEIGHT)

        self.edit.installEventFilter(self.filter)
        self.edit.setPlaceholderText(help_text if help_text else '')

        self.list.setMinimumWidth(self._MIN_WIDTH)
        self.list.setItemDelegate(SwitcherDelegate(self))
        self.list.setFocusPolicy(Qt.NoFocus)
        self.list.setSelectionBehavior(
            QAbstractItemView.SelectionBehavior.SelectItems
        )
        self.list.setSelectionMode(
            QAbstractItemView.SelectionMode.SingleSelection
        )
        self.list.setVerticalScrollMode(QAbstractItemView.ScrollPerItem)
        self.proxy.setSourceModel(self.model)
        self.list.setModel(self.proxy)

        # Layout
        layout = QVBoxLayout()
        layout.addWidget(self.edit)
        layout.addWidget(self.list)
        self.setLayout(layout)

        # Signals
        self.filter.sig_up_key_pressed.connect(self.previous_row)
        self.filter.sig_down_key_pressed.connect(self.next_row)
        self.filter.sig_enter_key_pressed.connect(self.enter)

        self.edit.textChanged.connect(self._on_search_text_changed)
        self.edit.returnPressed.connect(self.enter)

        self.list.clicked.connect(self.enter)
        self.list.clicked.connect(self.edit.setFocus)
        self.list.selectionModel().currentChanged.connect(
            self.current_item_changed)

        # Gives focus to text edit
        self.edit.setFocus()

    # ---- Helper methods
    # -------------------------------------------------------------------------
    def _add_item(self, item, last_item=True):
        """Perform common actions when adding items."""
        item.set_width(self._ITEM_WIDTH)
        self.model.appendRow(item)

        if last_item:
            # Only set the current row to the first item when the added item is
            # the last one in order to prevent performance issues when
            # adding multiple items
            self.set_current_row(0)
            self.set_height()

    # ---- API
    # -------------------------------------------------------------------------
    def clear(self):
        """Remove all items from the list and clear the search text."""
        self.set_placeholder_text('')
        self.model.clear()
        self.setMinimumHeight(self._MIN_HEIGHT)

    def set_placeholder_text(self, text):
        """Set the text appearing on the empty line edit."""
        self.edit.setPlaceholderText(text)

    def add_mode(self, token, description):
        """Add mode by token key and description."""
        if len(token) == 1:
            self._modes[token] = description
        else:
            raise Exception('Token must be of length 1!')

    def get_mode(self):
        """Get the current mode the switcher is in."""
        return self._mode_on

    def remove_mode(self, token):
        """Remove mode by token key."""
        if token in self._modes:
            self._modes.pop(token)

    def clear_modes(self):
        """Delete all modes spreviously defined."""
        del self._modes
        self._modes = {}

    def add_item(self, icon=None, title=None, description=None, shortcut=None,
                 section=None, data=None, tool_tip=None, action_item=False,
                 last_item=True, score=-1, use_score=True):
        """Add switcher list item."""
        item = SwitcherItem(
            parent=self.list,
            icon=icon,
            title=title,
            description=description,
            data=data,
            shortcut=shortcut,
            section=section,
            action_item=action_item,
            tool_tip=tool_tip,
            styles=self._item_styles,
            score=score,
            use_score=use_score
        )
        self._add_item(item, last_item=last_item)

    def add_separator(self):
        """Add separator item."""
        item = SwitcherSeparatorItem(
            parent=self.list, styles=self._item_separator_styles
        )
        self._add_item(item)

    def setup(self):
        """Setup list widget content based on filtering."""
        search_text = self.search_text_without_mode()

        # Build default view
        if self.search_text() == '':
            self._mode_on = ''
            self.clear()
            self.proxy.set_filter_by_score(False)
            self.sig_mode_selected.emit(self._mode_on)

            # This is necessary to show the Editor items first when results
            # come back from the Editor and Projects.
            self.proxy.sortBy('_score')

            # Show sections
            self.setup_sections()

            # Give focus to the first row
            self.set_current_row(0)

            return

        # Filter by text
        titles = []
        for row in range(self.model.rowCount()):
            item = self.model.item(row)
            if isinstance(item, SwitcherItem):
                title = item.get_title()
            else:
                title = ''

            titles.append(title)

        search_text = clean_string(search_text)
        scores = get_search_scores(
            str(search_text),
            titles,
            template=u"<b>{0}</b>"
        )

        for idx, (title, rich_title, score_value) in enumerate(scores):
            item = self.model.item(idx)
            if not self._is_separator(item) and not item.is_action_item():
                rich_title = rich_title.replace(" ", "&nbsp;")
                item.set_rich_title(rich_title)

            item.set_score(score_value)

        self.proxy.set_filter_by_score(True)
        self.proxy.sortBy('_score')

        # Graphical setup
        self.setup_sections()

        if self.count():
            self.set_current_row(0)
        else:
            self.set_current_row(-1)

        self.set_height()

    def setup_sections(self):
        """Setup which sections appear on the item list."""
        sections = []
        search_text = self.search_text_without_mode()

        for row in range(self.model.rowCount()):
            item_row = row

            # When there is search_text, we need to use the proxy model to get
            # the actual item's row.
            if search_text:
                model_index = self.proxy.mapToSource(self.proxy.index(row, 0))
                item_row = model_index.row()

            # Get item
            item = self.model.item(item_row)

            # When searching gives no result, the mapped items are None
            if item is None:
                continue

            # Get item section
            if isinstance(item, SwitcherItem):
                sections.append(item.get_section())
            else:
                sections.append('')

            # Decide if we need to make the item's section visible
            if row != 0:
                visible = sections[row] != sections[row - 1]
                if not self._is_separator(item):
                    item.set_section_visible(visible)
            else:
                # We need to remove this when a mode has several sections
                if not self._mode_on:
                    item.set_section_visible(True)

    def remove_section(self, section):
        """Remove all items in a section of the switcher."""
        # As we are removing items from the model, we need to iterate backwards
        # so that the indexes are not affected.
        for row in range(self.model.rowCount() - 1, -1, -1):
            item = self.model.item(row)
            if isinstance(item, SwitcherItem):
                if item._section == section:
                    self.model.removeRow(row)
                    continue

    def set_height(self):
        """Set height taking into account the number of items."""
        if self.count() >= self._MAX_NUM_ITEMS:
            switcher_height = self._MAX_HEIGHT
        elif self.count() != 0 and self.current_item():
            current_item = self.current_item()
            item_height = current_item.get_height()
            list_height = item_height * (self.count() + 3)
            edit_height = self.edit.height()
            spacing_height = self.layout().spacing() * 4
            switcher_height = list_height + edit_height + spacing_height
            switcher_height = max(switcher_height, self._MIN_HEIGHT)
        else:
            switcher_height = self._MIN_HEIGHT
        self.setFixedHeight(int(switcher_height))

    def set_position(self, top, parent=None):
        """Set the position of the dialog."""
        if parent is None:
            parent = self.parent()

        if parent is not None:
            geo = parent.geometry()
            width = self.list.width()  # This has been set in setup
            left = parent.geometry().width()/2 - width/2

            while parent:
                geo = parent.geometry()
                top += geo.top()
                left += geo.left()
                parent = parent.parent()

            self.move(round(left), top)

    @Slot(QModelIndex, QModelIndex)
    def current_item_changed(self, current, previous):
        """Handle item selection."""
        self.sig_item_changed.emit(self.current_item())

    # ---- Qt overrides
    # -------------------------------------------------------------------------
    @Slot()
    @Slot(QListWidgetItem)
    def enter(self, itemClicked=None):
        """Override Qt method."""
        row = self.current_row()
        model_index = self.proxy.mapToSource(self.proxy.index(row, 0))
        item = self.model.item(model_index.row())
        if item:
            mode = self._mode_on
            self.sig_item_selected.emit(
                item, mode, self.search_text_without_mode()
            )

    def reject(self):
        """Override Qt method."""
        # This prevents calling _on_search_text_changed, which unnecessarily
        # tries to populate the switcher when we're closing it.
        with signals_blocked(self.edit):
            self.set_search_text('')

        # Reset mode
        self._mode_on = ""

        self.sig_rejected.emit()
        super().reject()

    # ---- Helper methods: Lineedit widget
    # -------------------------------------------------------------------------
    def search_text(self):
        """Get the normalized (lowecase) content of the search text."""
        return str(self.edit.text()).lower()

    def search_text_without_mode(self):
        """Get search text without mode."""
        mode = self._mode_on
        if mode:
            search_text = self.search_text()[len(mode):]
        else:
            search_text = self.search_text()

        return search_text

    def set_search_text(self, string):
        """Set the content of the search text."""
        self.edit.setText(string)

    @qdebounced(timeout=250)
    def _on_search_text_changed(self):
        """Actions to take when the search text has changed."""
        if self.search_text() != "":
            search_text = self.search_text_without_mode()

            # Inform if mode has changed
            for key in self._modes:
                if search_text.startswith(key) and not self._mode_on:
                    self._mode_on = key
                    self.sig_mode_selected.emit(key)
                    break

            # Emit this signal only for the files mode for now. We'll see if
            # it's necessary for other modes later.
            if self._mode_on == "":
                self.sig_search_text_available.emit(clean_string(search_text))
            else:
                self.setup()
        else:
            self.setup()

    # ---- Helper methods: List widget
    # -------------------------------------------------------------------------
    def _is_separator(self, item):
        """Check if item is an separator item (SwitcherSeparatorItem)."""
        return isinstance(item, SwitcherSeparatorItem)

    def _select_row(self, steps):
        """Select row in list widget based on a number of steps with direction.

        Steps can be positive (next rows) or negative (previous rows).
        """
        row = self.current_row() + steps
        if 0 <= row < self.count():
            self.set_current_row(row)

    def count(self):
        """Get the item count in the list widget."""
        return self.proxy.rowCount()

    def current_row(self):
        """Return the current selected row in the list widget."""
        return self.list.currentIndex().row()

    def current_item(self):
        """Return the current selected item in the list widget."""
        row = self.current_row()
        model_index = self.proxy.mapToSource(self.proxy.index(row, 0))
        item = self.model.item(model_index.row())
        return item

    def set_current_row(self, row):
        """Set the current selected row in the list widget."""
        proxy_index = self.proxy.index(row, 0)
        selection_model = self.list.selectionModel()

        # https://doc.qt.io/qt-5/qitemselectionmodel.html#SelectionFlag-enum
        selection_model.setCurrentIndex(
            proxy_index, QItemSelectionModel.SelectionFlag.ClearAndSelect
        )

        # Ensure that the selected item is visible
        self.list.scrollTo(proxy_index, QAbstractItemView.EnsureVisible)

    def previous_row(self):
        """Select previous row in list widget."""
        steps = 1
        prev_row = self.current_row() - steps

        if prev_row == -1:
            self.set_current_row(self.count() - 1)
        else:
            if prev_row >= 0:
                # Need to map the filtered list to the actual model items
                list_index = self.proxy.index(prev_row, 0)
                model_index = self.proxy.mapToSource(list_index)
                item = self.model.item(model_index.row(), 0)
                if self._is_separator(item):
                    steps += 1
            self._select_row(-steps)

    def next_row(self):
        """Select next row in list widget."""
        steps = 1
        next_row = self.current_row() + steps

        # Need to map the filtered list to the actual model items
        list_index = self.proxy.index(next_row, 0)
        model_index = self.proxy.mapToSource(list_index)
        item = self.model.item(model_index.row(), 0)

        if next_row >= self.count():
            self.set_current_row(0)
        else:
            if item:
                if self._is_separator(item):
                    steps += 1
            self._select_row(steps)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import logging
from typing import Optional

from asyncssh import SSHClientConnection, SSHClient
from asyncssh.auth import PasswordChangeResponse
from asyncssh.public_key import KeyPairListArg


_logger = logging.getLogger(__name__)


class SpyderSSHClient(SSHClient):
    def __init__(self, client):
        self.client = client

    def connection_made(self, conn: SSHClientConnection) -> None:
        """Called when a connection is made

        This method is called as soon as the TCP connection completes.
        The `conn` parameter should be stored if needed for later use.

        :param conn:
            The connection which was successfully opened
        :type conn: :class:`SSHClientConnection`

        """

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Called when a connection is lost or closed

        This method is called when a connection is closed. If the
        connection is shut down cleanly, *exc* will be `None`.
        Otherwise, it will be an exception explaining the reason for
        the disconnect.

        :param exc:
            The exception which caused the connection to close, or
            `None` if the connection closed cleanly
        :type exc: :class:`Exception`

        """
        self.client._handle_connection_lost(exc)

    def debug_msg_received(
        self, msg: str, lang: str, always_display: bool
    ) -> None:
        """A debug message was received on this connection

        This method is called when the other end of the connection sends
        a debug message. Applications should implement this method if
        they wish to process these debug messages.

        :param msg:
            The debug message sent
        :param lang:
            The language the message is in
        :param always_display:
            Whether or not to display the message
        :type msg: `str`
        :type lang: `str`
        :type always_display: `bool`

        """
        _logger.debug(f"Debug message received: {msg}")

    def auth_completed(self) -> None:
        """Authentication was completed successfully

        This method is called when authentication has completed
        successfully. Applications may use this method to create
        whatever client sessions and direct TCP/IP or UNIX domain
        connections are needed and/or set up listeners for incoming
        TCP/IP or UNIX domain connections coming from the server.
        However, :func:`create_connection` now blocks until
        authentication is complete, so any code which wishes to
        use the SSH connection can simply follow that call and
        doesn't need to be performed in a callback.

        """

    def public_key_auth_requested(self) -> Optional[KeyPairListArg]:
        """Public key authentication has been requested

        This method should return a private key corresponding to
        the user that authentication is being attempted for.

        This method may be called multiple times and can return a
        different key to try each time it is called. When there are
        no keys left to try, it should return `None` to indicate
        that some other authentication method should be tried.

        If client keys were provided when the connection was opened,
        they will be tried before this method is called.

        If blocking operations need to be performed to determine the
        key to authenticate with, this method may be defined as a
        coroutine.

        :returns: A key as described in :ref:`SpecifyingPrivateKeys`
                  or `None` to move on to another authentication
                  method

        """

        return None  # pragma: no cover

    def password_auth_requested(self) -> Optional[str]:
        """Password authentication has been requested

        This method should return a string containing the password
        corresponding to the user that authentication is being
        attempted for. It may be called multiple times and can
        return a different password to try each time, but most
        servers have a limit on the number of attempts allowed.
        When there's no password left to try, this method should
        return `None` to indicate that some other authentication
        method should be tried.

        If a password was provided when the connection was opened,
        it will be tried before this method is called.

        If blocking operations need to be performed to determine the
        password to authenticate with, this method may be defined as
        a coroutine.

        :returns: A string containing the password to authenticate
                  with or `None` to move on to another authentication
                  method

        """

        return None  # pragma: no cover

    def password_change_requested(
        self, prompt: str, lang: str
    ) -> PasswordChangeResponse:
        """A password change has been requested

        This method is called when password authentication was
        attempted and the user's password was expired on the
        server. To request a password change, this method should
        return a tuple or two strings containing the old and new
        passwords. Otherwise, it should return `NotImplemented`.

        If blocking operations need to be performed to determine the
        passwords to authenticate with, this method may be defined
        as a coroutine.

        By default, this method returns `NotImplemented`.

        :param prompt:
            The prompt requesting that the user enter a new password
        :param lang:
            The language that the prompt is in
        :type prompt: `str`
        :type lang: `str`

        :returns: A tuple of two strings containing the old and new
                  passwords or `NotImplemented` if password changes
                  aren't supported

        """

        return NotImplemented  # pragma: no cover

    def password_changed(self) -> None:
        """The requested password change was successful

        This method is called to indicate that a requested password
        change was successful. It is generally followed by a call to
        :meth:`auth_completed` since this means authentication was
        also successful.

        """

    def password_change_failed(self) -> None:
        """The requested password change has failed

        This method is called to indicate that a requested password
        change failed, generally because the requested new password
        doesn't meet the password criteria on the remote system.
        After this method is called, other forms of authentication
        will automatically be attempted.

        """
